/*
 * Copyright (C) 1998,1999
 *	SHUDO Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

#include "compiler.h"

#include "opcodes.h"
#include "code.h"

#ifdef JITDO
#include "jitdo/jitdo.h"
#include "jitdo/NET_shudo_jitdo_Proxy_old.h"	/* for type H...Proxy */
#include "jitdo/NET_shudo_jitdo_VMAddress_old.h"/* for type H...VMAddress */
#endif	/* JITDO */


#define COMPILEDCODE(DST) \
    __asm__("movl  %0,%" #DST : : "m" (mb));	/* DST = mb */\
    METHOD_COMPILEDCODE(DST,DST);	/* DST = mb->CompiledCode */


volatile void assembledCode(
	JHandle *o /* 8(%ebp) */ , struct methodblock *mb /* 12(%ebp) */,
	int args_size, ExecEnv *ee
#ifdef RUNTIME_DEBUG
	, int runtime_debug
#endif
) {
  int32_t bytepcoff;	/* -4(%ebp) */
	/* to handle exceptions */
  /*
    %esi: stack_item *vars
   */

  /*
   * utilities
   */

  /* fill_cache */
#ifdef RUNTIME_DEBUG
#  define FILL_CACHE_DEBUG1(OPTOP1_REG, OPTOP2_REG) \
  if (runtime_debug) {\
    DEBUG_IN;\
    __asm__("pushl " #OPTOP2_REG "\n\tflds  (%esp)\n\t"\
	    "subl  $4,%esp\n\tfstpl (%esp)\n\t"\
	    "pushl " #OPTOP2_REG "\n\tpushl " #OPTOP2_REG);\
    __asm__("pushl " #OPTOP1_REG "\n\tflds  (%esp)\n\t"\
	    "subl  $4,%esp\n\tfstpl (%esp)\n\t"\
	    "pushl " #OPTOP1_REG "\n\tpushl " #OPTOP1_REG);\
    PUSH_CONSTSTR("  optop[-1]: 0x%08x %d %g, optop[-2]: 0x%08x %d %g\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $36,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#else
#  define FILL_CACHE_DEBUG1(OPTOP1_REG, OPTOP2_REG)
#endif

  CODE(opc_fill_cache, fill_cache, ST0, ST2, 0) {
    __asm__("popl  %ecx\n\tpopl  %edx");
    FILL_CACHE_DEBUG1(%ecx, %edx);
  }
  CODE(opc_fill_cache, fill_cache, ST1, ST4, 0) {
    __asm__("popl  %ecx");
    FILL_CACHE_DEBUG1(%edx, %ecx);
  }
  CODE(opc_fill_cache, fill_cache, ST2, ST2, 0) {
    FILL_CACHE_DEBUG1(%ecx, %edx);
  }
  CODE(opc_fill_cache, fill_cache, ST3, ST2, 0) {
    __asm__("popl  %edx");
    FILL_CACHE_DEBUG1(%ecx, %edx);
  }
  CODE(opc_fill_cache, fill_cache, ST4, ST4, 0) {
    FILL_CACHE_DEBUG1(%edx, %ecx);
  }

  /* to another state */
  CODE(opc_stateto0, stateto00, ST0, ST0, 0) {}
  CODE(opc_stateto0, stateto10, ST1, ST0, 0) { __asm__("pushl %edx"); }
  CODE(opc_stateto0, stateto20, ST2, ST0, 0) {
    __asm__("pushl %edx\n\t"
	    "pushl %ecx"); }
  CODE(opc_stateto0, stateto30, ST3, ST0, 0) { __asm__("pushl %ecx"); }
  CODE(opc_stateto0, stateto40, ST4, ST0, 0) {
    __asm__("pushl %ecx\n\t"
	    "pushl %edx"); }
  CODE(opc_stateto1, stateto01, ST0, ST1, 0) { __asm__("popl  %edx"); }
  CODE(opc_stateto1, stateto11, ST1, ST1, 0) {}
  CODE(opc_stateto1, stateto21, ST2, ST1, 0) {
    __asm__("pushl %edx\n\t"
	    "movl  %ecx,%edx"); }
  CODE(opc_stateto1, stateto31, ST3, ST1, 0) { __asm__("movl  %ecx,%edx"); }
  CODE(opc_stateto1, stateto41, ST4, ST1, 0) { __asm__("pushl %ecx"); }
  CODE(opc_stateto2, stateto02, ST0, ST2, 0) {
    __asm__("popl  %ecx\n\t"
	    "popl  %edx"); }
  CODE(opc_stateto2, stateto12, ST1, ST2, 0) {
    __asm__("movl  %edx,%ecx\n\t"
	    "popl  %edx"); }
  CODE(opc_stateto2, stateto22, ST2, ST2, 0) {}
  CODE(opc_stateto2, stateto32, ST3, ST2, 0) { __asm__("popl  %edx"); }
  CODE(opc_stateto2, stateto42, ST4, ST2, 0) { __asm__("xchg  %edx,%ecx"); }
  CODE(opc_stateto3, stateto03, ST0, ST3, 0) { __asm__("popl  %ecx"); }
  CODE(opc_stateto3, stateto13, ST1, ST3, 0) { __asm__("movl  %edx,%ecx"); }
  CODE(opc_stateto3, stateto23, ST2, ST3, 0) { __asm__("pushl %edx"); }
  CODE(opc_stateto3, stateto33, ST3, ST3, 0) {}
  CODE(opc_stateto3, stateto43, ST4, ST3, 0) {
    __asm__("pushl %ecx\n\t"
	    "movl  %edx,%ecx"); }
  CODE(opc_stateto4, stateto04, ST0, ST4, 0) {
    __asm__("popl  %edx\n\t"
	    "popl  %ecx"); }
  CODE(opc_stateto4, stateto14, ST1, ST4, 0) { __asm__("popl  %ecx"); }
  CODE(opc_stateto4, stateto24, ST2, ST4, 0) { __asm__("xchg  %edx,%ecx"); }
  CODE(opc_stateto4, stateto34, ST3, ST4, 0) {
    __asm__("movl  %ecx,%edx\n\t"
	    "popl  %ecx"); }
  CODE(opc_stateto4, stateto44, ST4, ST4, 0) {}

  /* to another state and jump */
#define JUMP_TO_ACCUMULATOR	__asm__("jmp   *%eax")

  CODE(opc_goto_st0, goto_st00, ST0, ST0, 0) {
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st0, goto_st10, ST1, ST0, 0) {
    __asm__("pushl %edx");
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st0, goto_st20, ST2, ST0, 0) {
    __asm__("pushl %edx\n\t"
	    "pushl %ecx");
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st0, goto_st30, ST3, ST0, 0) {
    __asm__("pushl %ecx");
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st0, goto_st40, ST4, ST0, 0) {
    __asm__("pushl %ecx\n\t"
	    "pushl %edx");
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st1, goto_st01, ST0, ST1, 0) {
    __asm__("popl  %edx");
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st1, goto_st11, ST1, ST1, 0) {
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st1, goto_st21, ST2, ST1, 0) {
    __asm__("pushl %edx\n\t"
	    "movl  %ecx,%edx");
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st1, goto_st31, ST3, ST1, 0) {
    __asm__("movl  %ecx,%edx");
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st1, goto_st41, ST4, ST1, 0) {
    __asm__("pushl %ecx");
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st2, goto_st02, ST0, ST2, 0) {
    __asm__("popl  %ecx\n\t"
	    "popl  %edx");
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st2, goto_st12, ST1, ST2, 0) {
    __asm__("movl  %edx,%ecx\n\t"
	    "popl  %edx");
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st2, goto_st22, ST2, ST2, 0) {
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st2, goto_st32, ST3, ST2, 0) {
    __asm__("popl  %edx");
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st2, goto_st42, ST4, ST2, 0) {
    __asm__("xchg  %edx,%ecx");
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st3, goto_st03, ST0, ST3, 0) {
    __asm__("popl  %ecx");
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st3, goto_st13, ST1, ST3, 0) {
    __asm__("movl  %edx,%ecx");
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st3, goto_st23, ST2, ST3, 0) {
    __asm__("pushl %edx");
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st3, goto_st33, ST3, ST3, 0) {
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st3, goto_st43, ST4, ST3, 0) {
    __asm__("pushl %ecx\n\t"
	    "movl  %edx,%ecx");
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st4, goto_st04, ST0, ST4, 0) {
    __asm__("popl  %edx\n\t"
	    "popl  %ecx");
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st4, goto_st14, ST1, ST4, 0) {
    __asm__("popl  %ecx");
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st4, goto_st24, ST2, ST4, 0) {
    __asm__("xchg  %edx,%ecx");
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st4, goto_st34, ST3, ST4, 0) {
    __asm__("movl  %ecx,%edx\n\t"
	    "popl  %ecx");
    JUMP_TO_ACCUMULATOR; }
  CODE(opc_goto_st4, goto_st44, ST4, ST4, 0) {
    JUMP_TO_ACCUMULATOR; }


  /*
   * invocation related codes
   */

  /* method head */
  CODE_WITHOUT_DEBUG(opc_methodhead, head, STANY, STSTA, 0) {
    __asm__("pushl %ebp\n\t"
	    "movl  %esp,%ebp\n\t"
	    "subl  $4,%esp\n\t"	/* 4 * #local_var */
#define SAVED_REG_AREA	16	/* 16 bytes: ecx,edx,edi,esi */
	    "pushl %ecx\n\t"
	    "pushl %edx\n\t"
	    "pushl %edi\n\t"
	    "pushl %esi");

#ifdef RUNTIME_DEBUG
    /* write class and method name */
    if (runtime_debug) {
      register struct methodblock *mb_reg asm("edi");

      DEBUG_IN;
      mb_reg = mb;

      __asm__("movl  " METHOD_NAME(%edi) ",%eax\n\t"
	      "testl %eax,%eax\n\t"
	      "jz    methodhead_name_done\n\t"
	      "pushl %eax");
      METHOD_CLAZZ(%edi, %eax);
      CB_NAME(%eax, %eax);	/* break %edi */
      __asm__("pushl %eax");
      PUSH_CONSTSTR("  %s#%s\n");
      __asm__("call  " SYMBOL(printf) "@PLT\n\t"
	      "addl  $12,%esp");
      FFLUSH;
      __asm__("methodhead_name_done:");
      DEBUG_OUT;
    }
#endif	/* RUNTIME_DEBUG */

#ifdef RUNTIME_DEBUG
    if (runtime_debug) {
      __asm__("pushl %eax\n\t"
	      "movl  %esp,%eax");
      DEBUG_IN;
      __asm__("pushl %eax");
      PUSH_CONSTSTR("  ESP: 0x%08x\n");
      __asm__("call  " SYMBOL(printf) "@PLT\n\t"
	      "addl  $8,%esp");
      FFLUSH;
      DEBUG_OUT;
      __asm__("popl  %eax");
    }
#endif

    /* %esi = ee->current_frame->vars */
    {
      register ExecEnv *cur_ee asm("eax");
      cur_ee = ee;
    }
    __asm__("movl  " EE_CURRENTFRAME(%eax) ",%edi\n\t"
	    "movl  " FRAME_VARS(%edi) ",%esi");
#ifdef RUNTIME_DEBUG
    if (runtime_debug) {
      DEBUG_IN;
      __asm__("pushl %esi");
      PUSH_CONSTSTR("  cached vars: 0x%08x\n");
      __asm__("call  " SYMBOL(printf) "@PLT\n\t"
	      "addl  $8,%esp");
      FFLUSH;
      DEBUG_OUT;
    }
#endif
  }

  /* method tail */
  CODE(opc_methodtail, tail, STANY, STSTA, 0) {
    __asm__("leal  -20(%ebp),%esp");
		/* -4 * (#local_var + #registers_on_stack) */
    __asm__("popl  %esi\n\t"
	    "popl  %edi\n\t"
	    "popl  %edx\n\t"
	    "popl  %ecx\n\t"
	    "leave\n\t"
	    "ret");
  }

  /* exception handler */
  CODE(opc_exception_handler, exception_handler, STANY, STSTA, 0) {
#ifdef RUNTIME_DEBUG
    __asm__("pushl %0" : : "m" (runtime_debug));
#endif
    __asm__("pushl %0\n\t"
	    "pushl %1\n\t"
	    "pushl %2"
	    : : "m" (bytepcoff), "m" (mb), "m" (ee));
    __asm__("call  " SYMBOL(searchCatchFrame) "@PLT");
#ifdef RUNTIME_DEBUG
    __asm__("addl  $16,%esp");
#else
    __asm__("addl  $12,%esp");
#endif
	/* eax is CatchFrame */

    __asm__("testl %eax,%eax\n\t"
	    "jz    exc_through");

    /* clear the stack */
    __asm__("leal  -" STR(SAVED_REG_AREA) "(%ebp),%esp");

    {	/* place the exception object on TOS */
      register ExecEnv *cur_ee asm("edi");
      register JHandle *excobj asm("edx");
      cur_ee = ee;
      excobj = cur_ee->exception.exc;
      __asm__("pushl %edx");
		/* now state 0 */

      cur_ee->exception.exc = NULL;
    }

    {
      register struct CatchFrame_w_state *cf asm("eax");
      register unsigned char *tgt asm("edi");
      register int32_t handler_state asm("ecx");

      handler_state = cf->state;
#ifdef RUNTIME_DEBUG
      if (runtime_debug) {
	DEBUG_IN;
	printf("  exc handler state: %d\n", handler_state);
	fflush(stdout);
	DEBUG_OUT;
      }
#endif

      /* shift to the state which handler assumes */
      __asm__("cmpl  $0,%ecx\n\t"	/* if (hanlder_state == 0) */
	      "je   exc_handler_shift_done\n\t"
	"exc_handler_shift_1:\n\t"
	      "cmpl  $1,%ecx\n\t"	/* if (handler_state == 1) */
	      "jne   exc_handler_shift_2\n\t"
	      "popl  %edx\n\t"
	      "jmp   exc_handler_shift_done\n\t"
	"exc_handler_shift_2:\n\t"
	      "cmpl  $3,%ecx\n\t"	/* if (handler_state == 3) */
	      "jne   exc_handler_shift_3\n\t"
	      "popl  %ecx\n\t"
	      "jmp   exc_handler_shift_done\n\t"
	"exc_handler_shift_3:\n\t"
	      "cmpl  $2,%ecx\n\t"	/* if (handler_stat == 2) */
	      "jne   exc_handler_shift_4\n\t"
	      "popl  %ecx\n\t"
	      "xorl  %edx,%edx\n\t"
	      "jmp   exc_handler_shift_done\n\t"
	"exc_handler_shift_4:\n\t"
	      "cmpl  $4,%ecx\n\t"	/* if (handler_stat == 4) */
	      "jne   exc_handler_shift_done\n\t"
	      "popl  %edx\n\t"
	      "xorl  %ecx,%ecx\n\t"
	"exc_handler_shift_done:");

      COMPILEDCODE(%edi);	/* tgt(edi) = mb->CompiledCode */
      tgt/*edi*/ += (int)cf/*eax*/->compiled_CatchFrame;
      __asm__("jmp   *%edi");
    }
    __asm__("exc_through:");
  }

  /* nop */
  CODE(opc_nop, nop, STANY, STSTA, 0) {
  }

  /* iconst_0 */
  CODE(opc_iconst_0, [ifa]const_0, ST0, ST1, 0) { __asm__("xorl  %edx,%edx"); }
  CODE(opc_iconst_0, [ifa]const_0, ST1, ST2, 0) { __asm__("xorl  %ecx,%ecx"); }
  CODE(opc_iconst_0, [ifa]const_0, ST2, ST4, 0) {
    __asm__("pushl %edx\n\t"
	    "xorl  %edx,%edx");
  }
  CODE(opc_iconst_0, [ifa]const_0, ST3, ST4, 0) { __asm__("xorl  %edx,%edx"); }
  CODE(opc_iconst_0, [ifa]const_0, ST4, ST2, 0) {
    __asm__("pushl %ecx\n\t"
	    "xorl  %ecx,%ecx");
  }

  /* iconst_m1, iconst_[1-5] */
#define CODE_ICONST_N(S, N) \
  CODE(opc_iconst_##S, iconst_##S, ST0, ST1, 0) {\
    __asm__("movl  $" #N ",%edx"); }\
  CODE(opc_iconst_##S, iconst_##S, ST1, ST2, 0) {\
    __asm__("movl  $" #N ",%ecx"); }\
  CODE(opc_iconst_##S, iconst_##S, ST2, ST4, 0){\
    __asm__("pushl %edx\n\t"\
	    "movl  $" #N ",%edx");\
  }\
  CODE(opc_iconst_##S, iconst_##S, ST3, ST4, 0) {\
    __asm__("movl  $" #N ",%edx"); }\
  CODE(opc_iconst_##S, iconst_##S, ST4, ST2, 0) {\
    __asm__("pushl %ecx\n\t"\
	    "movl  $" #N ",%ecx");\
  }

  CODE_ICONST_N(m1, -1);
  CODE_ICONST_N(1, 1);
  CODE_ICONST_N(2, 2);
  CODE_ICONST_N(3, 3);
  CODE_ICONST_N(4, 4);
  CODE_ICONST_N(5, 5);

  /* lconst_0 */
  CODE(opc_lconst_0, [ld]const_0, ST0, ST2, 0) {
    __asm__("xorl  %edx,%edx\n\t"
	    "xorl  %ecx,%ecx");
  }
  CODE(opc_lconst_0, [ld]const_0, ST1, ST2, 0) {
    __asm__("pushl %edx\n\t"
	    "xorl  %edx,%edx\n\t"
	    "xorl  %ecx,%ecx");
  }
  CODE(opc_lconst_0, [ld]const_0, ST2, ST2, 0) {
    __asm__("pushl %edx\n\t"
	    "pushl %ecx\n\t"
	    "xorl  %edx,%edx\n\t"
	    "xorl  %ecx,%ecx");
  }
  CODE(opc_lconst_0, [ld]const_0, ST3, ST2, 0) {
    __asm__("pushl %ecx\n\t"
	    "xorl  %edx,%edx\n\t"
	    "xorl  %ecx,%ecx");
  }
  CODE(opc_lconst_0, [ld]const_0, ST4, ST2, 0) {
    __asm__("pushl %ecx\n\t"
	    "pushl %edx\n\t"
	    "xorl  %ecx,%ecx\n\t"
	    "xorl  %edx,%edx");
  }

  /* lconst_1 */
#define LCONST_ST02 \
    __asm__("movl  $1,%ecx\n\t"\
	    "xorl  %edx,%edx")
	/* now state 2 */

#define LCONST_ST04 \
    __asm__("movl  $1,%edx\n\t"\
	    "xorl  %ecx,%ecx")
	/* now state 4 */

  CODE(opc_lconst_1, lconst_1, ST0, ST2, 0) {
    LCONST_ST02;
  }
  CODE(opc_lconst_1, lconst_1, ST1, ST2, 0) {
    __asm__("pushl %edx");	/* now state 0 */
    LCONST_ST02;
  }
  CODE(opc_lconst_1, lconst_1, ST2, ST4, 0) {
    __asm__("pushl %edx\n\t"
	    "pushl %ecx");	/* now state 0 */
    LCONST_ST04;
  }
  CODE(opc_lconst_1, lconst_1, ST3, ST4, 0) {
    __asm__("pushl %ecx");	/* now state 0 */
    LCONST_ST04;
  }
  CODE(opc_lconst_1, lconst_1, ST4, ST2, 0) {
    __asm__("pushl %ecx\n\t"
	    "pushl %edx");	/* now state 0 */
    LCONST_ST02;
  }

  /* fconst_1 */
  CODE(opc_fconst_1, fconst_1, ST0, ST1, 0) {
    __asm__("movl  $0x3f800000,%edx");
  }
  CODE(opc_fconst_1, fconst_1, ST1, ST2, 0) {
    __asm__("movl  $0x3f800000,%ecx");
  }
  CODE(opc_fconst_1, fconst_1, ST2, ST4, 0) {
    __asm__("pushl %edx\n\t"
	    "movl  $0x3f800000,%edx");
  }
  CODE(opc_fconst_1, fconst_1, ST3, ST4, 0) {
    __asm__("movl  $0x3f800000,%edx");
  }
  CODE(opc_fconst_1, fconst_1, ST4, ST2, 0) {
    __asm__("pushl %ecx\n\t"
	    "movl  $0x3f800000,%ecx");
  }

  /* fconst_2 */
  CODE(opc_fconst_2, fconst_2, ST0, ST1, 0) {
    __asm__("movl  $0x40000000,%edx");
  }
  CODE(opc_fconst_2, fconst_2, ST1, ST2, 0) {
    __asm__("movl  $0x40000000,%ecx");
  }
  CODE(opc_fconst_2, fconst_2, ST2, ST4, 0) {
    __asm__("pushl %edx\n\t"
	    "movl  $0x40000000,%edx");
  }
  CODE(opc_fconst_2, fconst_2, ST3, ST4, 0) {
    __asm__("movl  $0x40000000,%edx");
  }
  CODE(opc_fconst_2, fconst_2, ST4, ST2, 0) {
    __asm__("pushl %ecx\n\t"
	    "movl  $0x40000000,%ecx");
  }

  /* dconst_1 */
#define DCONST_ST02 \
    __asm__("movl  $0x3ff00000,%edx\n\t"	/* high word */\
	    "xorl  %ecx,%ecx")		/* low word */

#define DCONST_ST04 \
    __asm__("movl  $0x3ff00000,%ecx\n\t"	/* high word */\
	    "xorl  %edx,%edx")		/* low word */

  CODE(opc_dconst_1, dconst_1, ST0, ST2, 0) {
    DCONST_ST02;
  }
  CODE(opc_dconst_1, dconst_1, ST1, ST4, 0) {
    __asm__("pushl %edx");	/* now state 0 */
    DCONST_ST04;
  }
  CODE(opc_dconst_1, dconst_1, ST2, ST2, 0) {
    __asm__("pushl %edx\n\t"
	    "pushl %ecx");	/* now state 0 */
    DCONST_ST02;
  }
  CODE(opc_dconst_1, dconst_1, ST3, ST2, 0) {
    __asm__("pushl %ecx");	/* now state 0 */
    DCONST_ST02;
  }
  CODE(opc_dconst_1, dconst_1, ST4, ST4, 0) {
    __asm__("pushl %ecx\n\t"
	    "pushl %edx");	/* now state 0 */
    DCONST_ST04;
  }

  /* bipush */
#ifdef RUNTIME_DEBUG
#  define BIPUSH_DEBUG1(REG) \
  if (runtime_debug) {\
    DEBUG_IN;\
    __asm__("pushl " #REG "\n\t"\
	    "flds  (%esp)\n\t"\
	    "subl  $4,%esp\n\t"\
	    "fstpl (%esp)\n\t"\
	    "pushl " #REG "\n\t"\
	    "pushl " #REG);\
    PUSH_CONSTSTR("  0x%08x %d %g\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $20,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#else
#  define BIPUSH_DEBUG1(REG)
#endif
	/* const: value */
  CODE(opc_bipush, a_const, ST0, ST1, 0) {
    __asm__("movl  $" STR(CONST) ",%edx");
    BIPUSH_DEBUG1(%edx);
  }
  CODE(opc_bipush, a_const, ST1, ST2, 0) {
    __asm__("movl  $" STR(CONST) ",%ecx");
    BIPUSH_DEBUG1(%ecx);
  }
  CODE(opc_bipush, a_const, ST2, ST4, 0) {
    __asm__("pushl %edx\n\t"
	    "movl  $" STR(CONST) ",%edx");
    BIPUSH_DEBUG1(%edx);
  }
  CODE(opc_bipush, a_const, ST3, ST4, 0) {
    __asm__("movl  $" STR(CONST) ",%edx");
    BIPUSH_DEBUG1(%edx);
  }
  CODE(opc_bipush, a_const, ST4, ST2, 0) {
    __asm__("pushl %ecx\n\t"
	    "movl  $" STR(CONST) ",%ecx");
    BIPUSH_DEBUG1(%ecx);
  }

  /* ldc2_w */
	/* const: val[32:63], val[0:31] */
#ifdef RUNTIME_DEBUG
#  define LDC2_W_DEBUG1(OPTOP1_REG, OPTOP2_REG) \
  if (runtime_debug) {\
    DEBUG_IN;\
    __asm__("pushl " #OPTOP2_REG "\n\tpushl " #OPTOP1_REG "\n\t"\
	    "fldl  (%esp)\n\tfstpl (%esp)\n\t"\
	    "pushl " #OPTOP2_REG "\n\tpushl " #OPTOP1_REG "\n\t"\
	    "pushl " #OPTOP2_REG "\n\tpushl " #OPTOP1_REG);\
    PUSH_CONSTSTR("  0x%016llx %lld %g\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $28,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#else
#  define LDC2_W_DEBUG1(OPTOP1_REG, OPTOP2_REG)
#endif

  CODE(opc_ldc2_w, ldc2_w, ST0, ST2, 0) {
    __asm__("movl  $" STR(CONST) ",%edx\n\t"
	    "movl  $" STR(CONST) ",%ecx");
    LDC2_W_DEBUG1(%ecx, %edx);
  }
  CODE(opc_ldc2_w, ldc2_w, ST1, ST4, 0) {
    __asm__("pushl %edx\n\t"
	    "movl  $" STR(CONST) ",%ecx\n\t"
	    "movl  $" STR(CONST) ",%edx");
    LDC2_W_DEBUG1(%edx, %ecx);
  }
  CODE(opc_ldc2_w, ldc2_w, ST2, ST2, 0) {
    __asm__("pushl %edx\n\t"
	    "pushl %ecx\n\t"
	    "movl  $" STR(CONST) ",%edx\n\t"
	    "movl  $" STR(CONST) ",%ecx");
    LDC2_W_DEBUG1(%ecx, %edx);
  }
  CODE(opc_ldc2_w, ldc2_w, ST3, ST2, 0) {
    __asm__("pushl %ecx\n\t"
	    "movl  $" STR(CONST) ",%edx\n\t"
	    "movl  $" STR(CONST) ",%ecx");
    LDC2_W_DEBUG1(%ecx, %edx);
  }
  CODE(opc_ldc2_w, ldc2_w, ST4, ST4, 0) {
    __asm__("pushl %ecx\n\t"
	    "pushl %edx\n\t"
	    "movl  $" STR(CONST) ",%ecx\n\t"
	    "movl  $" STR(CONST) ",%edx");
    LDC2_W_DEBUG1(%edx, %ecx);
  }

  /* iload */
	/* const: index * 4 */
#ifdef RUNTIME_DEBUG
#  define ILOAD_DEBUG1(VAL) \
  if (runtime_debug) {\
    DEBUG_IN;\
    __asm__("pushl " #VAL "\n\t"\
	    "flds  (%esp)\n\t"\
	    "subl  $4,%esp\n\t"\
	    "fstpl (%esp)\n\t"\
	    "pushl " #VAL "\n\t"\
	    "pushl " #VAL "\n\t"\
	    "movl  $" STR(CONST) ",%eax\n\t"\
	    "sarl  $2,%eax\n\t"\
	    "negl  %eax\n\t"\
	    "pushl %eax");\
    PUSH_CONSTSTR("  var[%d]: 0x%08x %d %g\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $24,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#else
#  define ILOAD_DEBUG1(VAL)
#endif
  CODE(opc_iload, [ifa]load, ST0, ST1, 0) {
    __asm__("movl " STR(CONST) "(%esi),%edx");
    ILOAD_DEBUG1(%edx);
  }
  CODE(opc_iload, [ifa]load, ST1, ST2, 0) {
    __asm__("movl " STR(CONST) "(%esi),%ecx");
    ILOAD_DEBUG1(%ecx);
  }
  CODE(opc_iload, [ifa]load, ST2, ST4, 0) {
    __asm__("pushl %edx\n\t"
	    "movl  " STR(CONST) "(%esi),%edx");
    ILOAD_DEBUG1(%edx);
  }
  CODE(opc_iload, [ifa]load, ST3, ST4, 0) {
    __asm__("movl " STR(CONST) "(%esi),%edx");
    ILOAD_DEBUG1(%edx);
  }
  CODE(opc_iload, [ifa]load, ST4, ST2, 0) {
    __asm__("pushl %ecx\n\t"
	    "movl  " STR(CONST) "(%esi),%ecx");
    ILOAD_DEBUG1(%ecx);
  }

  /* lload */
	/* const: index * 4, (index + 1) * 4 */
#ifdef RUNTIME_DEBUG
#  define LLOAD_DEBUG1(OPTOP1_REG, OPTOP2_REG) \
  if (runtime_debug) {\
    DEBUG_IN;\
    __asm__("pushl " #OPTOP2_REG "\n\tpushl " #OPTOP1_REG "\n\t"\
	    "pushl " #OPTOP2_REG "\n\tpushl " #OPTOP1_REG "\n\t"\
	    "pushl " #OPTOP2_REG "\n\tpushl " #OPTOP1_REG "\n\t"\
	    "movl  $" STR(CONST) ",%eax\n\t"\
	    "sarl  $2,%eax\n\t"\
	    "negl  %eax\n\t"\
	    "subl  $1,%eax\n\t"\
	    "pushl %eax");\
    PUSH_CONSTSTR("  var[%d]: 0x%016llx, %lld, %g\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $32,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#else
#  define LLOAD_DEBUG1(OPTOP1_REG, OPTOP2_REG)
#endif

#define LLOAD_ST02 \
    __asm__("movl  " STR(CONST) "(%esi),%ecx\n\t"\
	    "movl  " STR(CONST) "(%esi),%edx");\
    LLOAD_DEBUG1(%ecx, %edx)
#define LLOAD_ST04 \
    __asm__("movl  " STR(CONST) "(%esi),%edx\n\t"\
	    "movl  " STR(CONST) "(%esi),%ecx");\
    LLOAD_DEBUG1(%edx, %ecx)

  CODE(opc_lload, [ld]load, ST0, ST2, 0) {
    LLOAD_ST02;
  }
  CODE(opc_lload, [ld]load, ST1, ST2, 0) {
    __asm__("pushl %edx");	/* now state 0 */
    LLOAD_ST02;
  }
  CODE(opc_lload, [ld]load, ST2, ST4, 0) {
    __asm__("pushl %edx\n\t"
	    "pushl %ecx");	/* now state 0 */
    LLOAD_ST04;
  }
  CODE(opc_lload, [ld]load, ST3, ST4, 0) {
    __asm__("pushl %ecx");	/* now state 0 */
    LLOAD_ST04;
  }
  CODE(opc_lload, [ld]load, ST4, ST2, 0) {
    __asm__("pushl %ecx\n\t"
	    "pushl %edx");	/* now state 0 */
    LLOAD_ST02;
  }

  /* iload_[0-3] */
#define CODE_ILOAD_N(N, OFF) \
  CODE(opc_iload_##N, [ifa]load_##N, ST0, ST1, 0) {\
    __asm__("movl  " #OFF "(%esi),%edx");\
    BIPUSH_DEBUG1(%edx);\
  }\
  CODE(opc_iload_##N, [ifa]load_##N, ST1, ST2, 0) {\
    __asm__("movl  " #OFF "(%esi),%ecx");\
    BIPUSH_DEBUG1(%ecx);\
  }\
  CODE(opc_iload_##N, [ifa]load_##N, ST2, ST4, 0) {\
    __asm__("pushl %edx\n\t"\
	    "movl  " #OFF "(%esi),%edx");\
    BIPUSH_DEBUG1(%edx);\
  }\
  CODE(opc_iload_##N, [ifa]load_##N, ST3, ST4, 0) {\
    __asm__("movl  " #OFF "(%esi),%edx");\
    BIPUSH_DEBUG1(%edx);\
  }\
  CODE(opc_iload_##N, [ifa]load_##N, ST4, ST2, 0) {\
    __asm__("pushl %ecx\n\t"\
	    "movl  " #OFF "(%esi),%ecx");\
    BIPUSH_DEBUG1(%ecx);\
  }

  CODE_ILOAD_N(0, );
  CODE_ILOAD_N(1, -4);
  CODE_ILOAD_N(2, -8);
  CODE_ILOAD_N(3, -12);

  /* lload_[0-3] */
#define LLOAD_N_ST02(OFF1, OFF2) \
    __asm__("movl  " #OFF1 "(%esi),%ecx\n\t"\
	    "movl  " #OFF2 "(%esi),%edx");\
    LDC2_W_DEBUG1(%ecx, %edx)
#define LLOAD_N_ST04(OFF1, OFF2) \
    __asm__("movl  " #OFF1 "(%esi),%edx\n\t"\
	    "movl  " #OFF2 "(%esi),%ecx");\
    LDC2_W_DEBUG1(%edx, %ecx)

#define CODE_LLOAD_N(N, OFF1, OFF2) \
  CODE(opc_lload_##N, [ld]load_##N, ST0, ST2, 0) {\
    LLOAD_N_ST02(OFF1, OFF2);\
  }\
  CODE(opc_lload_##N, [ld]load_##N, ST1, ST2, 0) {\
    __asm__("pushl %edx");	/* now state 0 */\
    LLOAD_N_ST02(OFF1, OFF2);\
  }\
  CODE(opc_lload_##N, [ld]load_##N, ST2, ST4, 0) {\
    __asm__("pushl %edx\n\t"\
	    "pushl %ecx");	/* now state 0 */\
    LLOAD_N_ST04(OFF1, OFF2);\
  }\
  CODE(opc_lload_##N, [ld]load_##N, ST3, ST4, 0) {\
    __asm__("pushl %ecx");	/* now state 0 */\
    LLOAD_N_ST04(OFF1, OFF2);\
  }\
  CODE(opc_lload_##N, [ld]load_##N, ST4, ST2, 0) {\
    __asm__("pushl %ecx\n\t"\
	    "pushl %edx");	/* now state 0 */\
    LLOAD_N_ST02(OFF1, OFF2);\
  }

  CODE_LLOAD_N(0, -4, );
  CODE_LLOAD_N(1, -8, -4);
  CODE_LLOAD_N(2, -12, -8);
  CODE_LLOAD_N(3, -16, -12);

  /*
   * array access
   */

  /* array_check */
#if 0
  CODE(opc_array_check, array_check, ST[24], ST[24], 1) {
    if (!h) { SIGNAL_ERROR0(NullPointerException); }

    if ((index < 0) || (index >= obj_length(h))) {
      SIGNAL_ERROR0(ArrayIndexOutOfBoundsException);
    }
  }
#endif

#ifdef RUNTIME_DEBUG
#  define ARRAY_CHECK_DEBUG1(INDEX) \
  if (runtime_debug) {\
    DEBUG_IN;\
    __asm__("pushl " #INDEX);\
    PUSH_CONSTSTR("  index:  %d\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $8,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#  define ARRAY_CHECK_DEBUG2(LEN) \
  if (runtime_debug) {\
    DEBUG_IN;\
    __asm__("pushl " #LEN);\
    PUSH_CONSTSTR("  length: %d\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $8,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#else
#  define ARRAY_CHECK_DEBUG1(INDEX)
#  define ARRAY_CHECK_DEBUG2(INDEX)
#endif

#ifndef NO_CHECK
#  define ARRAY_CHECK(HANDLE, INDEX, LABEL, STATE) \
    NULL_TEST(HANDLE, LABEL "_1", STATE);\
    \
    ARRAY_CHECK_DEBUG1(INDEX);\
    __asm__("testl " #INDEX "," #INDEX "\n\t"\
	    "jl    " LABEL "_exc");\
    OBJ_LENGTH(HANDLE, %edi);			/* edi = obj_length(handle) */\
    ARRAY_CHECK_DEBUG2(%edi);\
    __asm__("cmpl  %edi," #INDEX "\n\t"\
	    "jl    " LABEL "_done");\
    __asm__(LABEL "_exc:");		/* label */\
    SIGNAL_ERROR0(ArrayIndexOutOfBoundsException, STATE);\
    __asm__(LABEL "_done:")		/* label */
#else
#  define ARRAY_CHECK(HANDLE, INDEX, LABEL, STATE) \
    OBJ_LENGTH(HANDLE, %edi);			/* edi = obj_length(handle) */
#endif	/* NO_CHECK */
	/* destroy %edi */
	/* edi = obj_length(handle) */

  CODE(opc_array_check, array_check, ST2, ST2, 1) {
    ARRAY_CHECK(%edx, %ecx, "arychk_st2", 2);
  }
  CODE(opc_array_check, array_check, ST4, ST4, 1) {
    ARRAY_CHECK(%ecx, %edx, "arychk_st4", 4);
  }

  /* iaload */
	/* compile: fill_cache, array_check, iaload */
  CODE(opc_iaload, [ifa]aload, ST2, ST1, 0) {
    UNHAND(%edx, %eax);
    __asm__("movl  (%eax,%ecx,4),%edx");
    BIPUSH_DEBUG1(%edx);
  }
  CODE(opc_iaload, [ifa]aload, ST4, ST3, 0) {
    UNHAND(%ecx, %eax);
    __asm__("movl  (%eax,%edx,4),%ecx");
    BIPUSH_DEBUG1(%ecx);
  }

  /* laload */
	/* compile: fill_cache, array_check, laload */
  CODE(opc_laload, [ld]aload, ST2, ST4, 0) {
    UNHAND(%edx, %eax);
    __asm__("leal  (%eax,%ecx,8),%edi\n\t"
	    "movl  (%edi),%edx\n\t"
	    "movl  4(%edi),%ecx");
    LDC2_W_DEBUG1(%edx, %ecx);
  }
  CODE(opc_laload, [ld]aload, ST4, ST2, 0) {
    UNHAND(%ecx, %eax);
    __asm__("leal  (%eax,%edx,8),%edi\n\t"
	    "movl  (%edi),%ecx\n\t"
	    "movl  4(%edi),%edx");
    LDC2_W_DEBUG1(%ecx, %edx);
  }

  /* baload */
	/* compile: fill_cache, array_check, baload */
#if 0
  CODE(opc_baload, baload, ST2, ST1, 0) {
    register HArrayOfByte *h asm("edx");
    register int index asm("ecx");
    register ArrayOfByte *array asm("eax");

    array = unhand(h);
    edx.i = array->body[index];
  }
#endif
  CODE(opc_baload, baload, ST2, ST1, 0) {
    UNHAND(%edx, %eax);
    __asm__("movsbl (%eax,%ecx),%edx");
    BIPUSH_DEBUG1(%edx);
  }
  CODE(opc_baload, baload, ST4, ST3, 0) {
    UNHAND(%ecx, %eax);
    __asm__("movsbl (%eax,%edx),%ecx");
    BIPUSH_DEBUG1(%ecx);
  }

  /* caload */
	/* compile: fill_cache, array_check, caload */
#if 0
  CODE(opc_caload, caload, ST2, ST1, 0) {
    register HArrayOfChar *h asm("edx");
    register int index asm("ecx");
    register ArrayOfChar *array asm("eax");

    array = unhand(h);
    edx.i = array->body[index];
  }
#endif
  CODE(opc_caload, caload, ST2, ST1, 0) {
    UNHAND(%edx, %eax);
    __asm__("movzwl (%eax,%ecx,2),%edx");
    BIPUSH_DEBUG1(%edx);
  }
  CODE(opc_caload, caload, ST4, ST3, 0) {
    UNHAND(%ecx, %eax);
    __asm__("movzwl (%eax,%edx,2),%ecx");
    BIPUSH_DEBUG1(%ecx);
  }

  /* saload */
	/* compile: fill_cache, array_check, saload */
#if 0
  CODE(opc_saload, saload, ST2, ST1, 0) {
    register HArrayOfShort *h asm("edx");
    register int index asm("ecx");
    register ArrayOfShort *array asm("eax");

    array = unhand(h);
    edx.i = array->body[index];
  }
#endif
  CODE(opc_saload, saload, ST2, ST1, 0) {
    UNHAND(%edx, %eax);
    __asm__("movswl (%eax,%ecx,2),%edx");
    BIPUSH_DEBUG1(%edx);
  }
  CODE(opc_saload, saload, ST4, ST3, 0) {
    UNHAND(%ecx, %eax);
    __asm__("movswl (%eax,%edx,2),%ecx");
    BIPUSH_DEBUG1(%ecx);
  }

  /* istore */
	/* const: index * 4 */
  CODE(opc_istore, [ifa]store, ST0, ST0, 0) {
    __asm__("popl  %edx\n\t"
	    "movl  %edx," STR(CONST) "(%esi)");
    ILOAD_DEBUG1(%edx);
  }
  CODE(opc_istore, [ifa]store, ST1, ST0, 0) {
    __asm__("movl  %edx," STR(CONST) "(%esi)");
    ILOAD_DEBUG1(%edx);
  }
  CODE(opc_istore, [ifa]store, ST2, ST1, 0) {
    __asm__("movl  %ecx," STR(CONST) "(%esi)");
    ILOAD_DEBUG1(%ecx);
  }
  CODE(opc_istore, [ifa]store, ST3, ST0, 0) {
    __asm__("movl  %ecx," STR(CONST) "(%esi)");
    ILOAD_DEBUG1(%ecx);
  }
  CODE(opc_istore, [ifa]store, ST4, ST3, 0) {
    __asm__("movl  %edx," STR(CONST) "(%esi)");
    ILOAD_DEBUG1(%edx);
  }

  /* lstore */
	/* const: index * 4, (index + 1) * 4 */
	/* compile: fill_cache, lstore */
  CODE(opc_lstore, [ld]store, ST2, ST0, 0) {
    __asm__("movl  %ecx," STR(CONST) "(%esi)\n\t"
	    "movl  %edx," STR(CONST) "(%esi)");
    LLOAD_DEBUG1(%ecx, %edx);
  }
  CODE(opc_lstore, [ld]store, ST4, ST0, 0) {
    __asm__("movl  %edx," STR(CONST) "(%esi)\n\t"
	    "movl  %ecx," STR(CONST) "(%esi)");
    LLOAD_DEBUG1(%edx, %ecx);
  }

  /* istore_[0-3] */
#define CODE_ISTORE_N(N, OFF) \
  CODE(opc_istore_##N, [ifa]store_##N, ST0, ST0, 0) {\
    __asm__("popl  %edx\n\t"		/* now state 1 */\
	    "movl  %edx," #OFF "(%esi)");\
    BIPUSH_DEBUG1(%edx);\
  }\
  CODE(opc_istore_##N, [ifa]store_##N, ST1, ST0, 0) {\
    __asm__("movl  %edx," #OFF "(%esi)");\
    BIPUSH_DEBUG1(%edx);\
  }\
  CODE(opc_istore_##N, [ifa]store_##N, ST2, ST1, 0) {\
    __asm__("movl  %ecx," #OFF "(%esi)");\
    BIPUSH_DEBUG1(%ecx);\
  }\
  CODE(opc_istore_##N, [ifa]store_##N, ST3, ST0, 0) {\
    __asm__("movl  %ecx," #OFF "(%esi)");\
    BIPUSH_DEBUG1(%ecx);\
  }\
  CODE(opc_istore_##N, [ifa]store_##N, ST4, ST3, 0) {\
    __asm__("movl  %edx," #OFF "(%esi)");\
    BIPUSH_DEBUG1(%edx);\
  }

  CODE_ISTORE_N(0, );
  CODE_ISTORE_N(1, -4);
  CODE_ISTORE_N(2, -8);
  CODE_ISTORE_N(3, -12);

  /* lstore_[0-3] */
#define LSTORE_N_ST2(OFF1, OFF2) \
    __asm__("movl  %ecx," #OFF1 "(%esi)\n\t"\
	    "movl  %edx," #OFF2 "(%esi)");\
    LDC2_W_DEBUG1(%ecx, %edx)
#define LSTORE_N_ST4(OFF1, OFF2) \
    __asm__("movl  %edx," #OFF1 "(%esi)\n\t"\
	    "movl  %ecx," #OFF2 "(%esi)");\
    LDC2_W_DEBUG1(%edx, %ecx)

#define CODE_LSTORE_N(N, OFF1, OFF2) \
  CODE(opc_lstore_##N, [ld]store_##N, ST0, ST0, 0) {\
    __asm__("popl  %ecx\n\t"\
	    "popl  %edx");		/* now state 2 */\
    LSTORE_N_ST2(OFF1, OFF2);\
  }\
  CODE(opc_lstore_##N, [ld]store_##N, ST1, ST0, 0) {\
    __asm__("popl  %ecx");		/* now state 4 */\
    LSTORE_N_ST4(OFF1, OFF2);\
  }\
  CODE(opc_lstore_##N, [ld]store_##N, ST2, ST0, 0) {\
    LSTORE_N_ST2(OFF1, OFF2);\
  }\
  CODE(opc_lstore_##N, [ld]store_##N, ST3, ST0, 0) {\
    __asm__("popl  %edx");		/* now state 2 */\
    LSTORE_N_ST2(OFF1, OFF2);\
  }\
  CODE(opc_lstore_##N, [ld]store_##N, ST4, ST0, 0) {\
    LSTORE_N_ST4(OFF1, OFF2);\
  }

  CODE_LSTORE_N(0, -4, );
  CODE_LSTORE_N(1, -8, -4);
  CODE_LSTORE_N(2, -12, -8);
  CODE_LSTORE_N(3, -16, -12);

  /* iastore */
	/* compile: iastore1, fill_cache, array_check, iastore */
  CODE(opc_iastore1, [ifa]astore1, ST0, ST0, 0) {
    __asm__("popl  %eax");
    BIPUSH_DEBUG1(%eax);
  }
  CODE(opc_iastore1, [ifa]astore1, ST1, ST0, 0) {
    __asm__("movl  %edx,%eax");
    BIPUSH_DEBUG1(%eax);
  }
  CODE(opc_iastore1, [ifa]astore1, ST2, ST1, 0) {
    __asm__("movl  %ecx,%eax");
    BIPUSH_DEBUG1(%eax);
  }
  CODE(opc_iastore1, [ifa]astore1, ST3, ST0, 0) {
    __asm__("movl  %ecx,%eax");
    BIPUSH_DEBUG1(%eax);
  }
  CODE(opc_iastore1, [ifa]astore1, ST4, ST3, 0) {
    __asm__("movl  %edx,%eax");
    BIPUSH_DEBUG1(%eax);
  }
	/* eax: value */

  CODE(opc_iastore, [ifa]astore, ST2, ST0, 0) {
    UNHAND(%edx, %edi);
    __asm__("movl  %eax,(%edi,%ecx,4)");	/* array->body[ecx] = eax */
  }
  CODE(opc_iastore, [ifa]astore, ST4, ST0, 0) {
    UNHAND(%ecx, %edi);
    __asm__("movl  %eax,(%edi,%edx,4)");	/* array->body[edx] = eax */
  }

  /* aastore */
	/* compile: iastore1, fill_cache, array_check, aastore */
  /*
   * assumption: eax is obj_handle(handle).
   */
#ifdef RUNTIME_DEBUG
#  define AASTORE_TEST_DEBUG \
  if (runtime_debug) {\
    DEBUG_IN;\
    __asm__("pushl %edi\n\tpushl %ecx\n\tpushl %edx");\
    PUSH_CONSTSTR("  edx: 0x%08x, ecx: 0x%08x, edi: 0x%08x\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $16,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#else
#  define AASTORE_TEST_DEBUG
#endif

#ifndef NO_CHECK
#  define AASTORE_TEST(OBJ, STATE) \
    __asm__("pushl %eax");\
    FUNCCALL_IN(2);\
    __asm__("pushl %0" : : "m" (ee));	/* push ee */\
    AASTORE_TEST_DEBUG;\
    __asm__("pushl (" #OBJ ",%edi,4)\n\t"\
				/* push array->body[obj_length(..)] */\
	    "pushl %eax\n\t"		/* push value */\
	/* call is_instance_of(value, array->body[eax], ee) */\
	    "call  " SYMBOL(is_instance_of) "@PLT\n\t"\
	    "addl  $12,%esp");\
	/* if (eax)  goto ... */\
    __asm__("testl %eax,%eax");\
    FUNCCALL_OUT(2);\
    __asm__("popl  %eax");\
    \
    __asm__("jnz   aastore_st" #STATE "_1");\
    SIGNAL_ERROR0(ArrayStoreException, STATE);\
    __asm__("aastore_st" #STATE "_1:")
#else
#  define AASTORE_TEST(OBJ, STATE)
#endif	/* NO_CHECK */

  CODE(opc_aastore, aastore, ST2, ST0, 0) {
    UNHAND(%edx, %edx);
    AASTORE_TEST(%edx, 2);
    __asm__("movl  %eax,(%edx,%ecx,4)");/* array->body[ecx] = eax */
  }
  CODE(opc_aastore, aastore, ST4, ST0, 0) {
    UNHAND(%ecx, %ecx);
    AASTORE_TEST(%ecx, 4);
    __asm__("movl  %eax,(%ecx,%edx,4)");/* array->body[ecx] = eax */
  }

  /* lastore */
	/* compile: fill_cache, lastore */
#define CODE_LASTORE(OPTOP1_REG, OPTOP2_REG, STATE) \
  CODE(opc_lastore, [ld]astore, ST##STATE, ST0, 1) {\
    __asm__("popl  %eax\n\t"\
	    "movl  (%esp),%edi");\
    ARRAY_CHECK(%edi, %eax, "lastore_st" #STATE, STATE);\
    __asm__("popl  %edi");\
    UNHAND(%edi, %edi);\
    __asm__("leal  (%edi,%eax,8),%edi\n\t"\
	    "movl  " #OPTOP1_REG ",(%edi)\n\t"\
	    "movl  " #OPTOP2_REG ",4(%edi)");\
    LDC2_W_DEBUG1(OPTOP1_REG, OPTOP2_REG);\
  }

  CODE_LASTORE(%ecx, %edx, 2);
  CODE_LASTORE(%edx, %ecx, 4);

  /* bastore */
	/* compile: iastore1, fill_cache, array_check, bastore */
  CODE(opc_bastore, bastore, ST2, ST0, 0) {
    UNHAND(%edx, %edi);
    __asm__("movb  %al,(%edi,%ecx)");
    BIPUSH_DEBUG1(%eax);
  }
  CODE(opc_bastore, bastore, ST4, ST0, 0) {
    UNHAND(%ecx, %edi);
    __asm__("movb  %al,(%edi,%edx)");
    BIPUSH_DEBUG1(%eax);
  }

  /* castore */
	/* compile: iastore1, fill_cache, array_check, castore */
  CODE(opc_castore, castore, ST2, ST0, 0) {
    UNHAND(%edx, %edi);
    __asm__("movw  %ax,(%edi,%ecx,2)");
    BIPUSH_DEBUG1(%eax);
  }
  CODE(opc_castore, castore, ST4, ST0, 0) {
    UNHAND(%ecx, %edi);
    __asm__("movw  %ax,(%edi,%edx,2)");
    BIPUSH_DEBUG1(%eax);
  }

  /* sastore */
	/* compile: iastore1, fill_cache, array_check, sastore */
  CODE(opc_sastore, sastore, ST2, ST0, 0) {
    UNHAND(%edx, %edi);
    __asm__("movw  %ax,(%edi,%ecx,2)");
    BIPUSH_DEBUG1(%eax);
  }
  CODE(opc_sastore, sastore, ST4, ST0, 0) {
    UNHAND(%ecx, %edi);
    __asm__("movw  %ax,(%edi,%edx,2)");
    BIPUSH_DEBUG1(%eax);
  }

  /* pop */
  CODE(opc_pop, pop, ST0, ST0, 0) {
    __asm__("addl  $4,%esp");
  }
  CODE(opc_pop, pop, ST1, ST0, 0) {}
  CODE(opc_pop, pop, ST2, ST1, 0) {}
  CODE(opc_pop, pop, ST3, ST0, 0) {}
  CODE(opc_pop, pop, ST4, ST3, 0) {}

  /* pop2 */
  CODE(opc_pop2, pop2, ST0, ST0, 0) {
    __asm__("addl  $8,%esp");
  }
  CODE(opc_pop2, pop2, ST1, ST0, 0) {
    __asm__("addl  $4,%esp");
  }
  CODE(opc_pop2, pop2, ST2, ST0, 0) {}
  CODE(opc_pop2, pop2, ST3, ST0, 0) {
    __asm__("addl  $4,%esp");
  }
  CODE(opc_pop2, pop2, ST4, ST0, 0) {}

  /* dup */
  CODE(opc_dup, dup, ST0, ST1, 0) {
    __asm__("movl  (%esp),%edx");
    BIPUSH_DEBUG1(%edx);
  }
  CODE(opc_dup, dup, ST1, ST2, 0) {
    __asm__("movl  %edx,%ecx");
    BIPUSH_DEBUG1(%ecx);
  }
  CODE(opc_dup, dup, ST2, ST2, 0) {
    __asm__("pushl %edx\n\t"
	    "movl  %ecx,%edx");
    BIPUSH_DEBUG1(%edx);
  }
  CODE(opc_dup, dup, ST3, ST2, 0) {
    __asm__("movl  %ecx,%edx");
    BIPUSH_DEBUG1(%edx);
  }
  CODE(opc_dup, dup, ST4, ST2, 0) {
    __asm__("pushl %ecx\n\t"
	    "movl  %edx,%ecx");
    BIPUSH_DEBUG1(%ecx);
  }

  /* dup_x1 */
  CODE(opc_dup_x1, dup_x1, ST0, ST2, 0) {
#if 0
    __asm__("popl  %ecx\n\t"		/* now state 3 */
	    "movl  (%esp),%edx\n\t"
	    "movl  %ecx,(%esp)");
#else
    __asm__("popl  %ecx\n\t"		/* now state 3 */
	    "popl  %edx\n\t"
	    "pushl %ecx");
#endif
  }
  CODE(opc_dup_x1, dup_x1, ST1, ST4, 0) {
#if 0
    __asm__("movl  (%esp),%ecx\n\t"
	    "movl  %edx,(%esp)");
#else
    __asm__("popl  %ecx\n\t"
	    "pushl %edx");
#endif
  }
  CODE(opc_dup_x1, dup_x1, ST2, ST2, 0) {
    __asm__("pushl %ecx");
  }
  CODE(opc_dup_x1, dup_x1, ST3, ST2, 0) {
#if 0
    __asm__("movl  (%esp),%edx\n\t"
	    "movl  %ecx,(%esp)");
#else
    __asm__("popl  %edx\n\t"
	    "pushl %ecx");
#endif
  }
  CODE(opc_dup_x1, dup_x1, ST4, ST4, 0) {
    __asm__("pushl %edx");
  }

  /* dup_x2 */
#if 0
#define DUP_X2_ST24(OPTOP1_REG) \
    __asm__("pushl (%esp)\n\t"\
	    "movl  " #OPTOP1_REG ",4(%esp)")
#else
#define DUP_X2_ST24(OPTOP1_REG) \
    __asm__("popl  %eax\n\t"\
	    "pushl " #OPTOP1_REG "\n\t"\
	    "pushl %eax")
#endif

  CODE(opc_dup_x2, dup_x2, ST0, ST2, 0) {
    __asm__("popl  %ecx\n\t"
	    "popl  %edx");		/* now state 2 */
    DUP_X2_ST24(%ecx);
  }
  CODE(opc_dup_x2, dup_x2, ST1, ST4, 0) {
    __asm__("popl  %ecx");		/* now state 4 */
    DUP_X2_ST24(%edx);
  }
  CODE(opc_dup_x2, dup_x2, ST2, ST2, 0) {
    DUP_X2_ST24(%ecx);
  }
  CODE(opc_dup_x2, dup_x2, ST3, ST2, 0) {
    __asm__("popl  %edx");		/* now state 2 */
    DUP_X2_ST24(%ecx);
  }
  CODE(opc_dup_x2, dup_x2, ST4, ST4, 0) {
    DUP_X2_ST24(%edx);
  }

  /* dup2 */
  CODE(opc_dup2, dup2, ST0, ST2, 0) {
#if 0
    __asm__("movl  (%esp),%ecx\n\t"
	    "movl  4(%esp),%edx");
#else
    __asm__("popl  %ecx\n\t"
	    "popl  %edx\n\t"
	    "subl  $8,%esp");
#endif
  }
  CODE(opc_dup2, dup2, ST1, ST4, 0) {
    __asm__("movl  (%esp),%ecx\n\t"
	    "pushl %edx");
  }
  CODE(opc_dup2, dup2, ST2, ST2, 0) {
    __asm__("pushl %edx\n\t"
	    "pushl %ecx");
  }
  CODE(opc_dup2, dup2, ST3, ST2, 0) {
    __asm__("movl  (%esp),%edx\n\t"
	    "pushl %ecx");
  }
  CODE(opc_dup2, dup2, ST4, ST4, 0) {
    __asm__("pushl %ecx\n\t"
	    "pushl %edx");
  }

  /* dup2_x1 */
#define DUP2_X1_ST24(OPTOP1_REG, OPTOP2_REG) \
    __asm__("popl  %eax\n\t"		/* eax = optop[-3] */\
	    "pushl " #OPTOP2_REG "\n\t"	/* optop[-5] = optop[-2] */\
	    "pushl " #OPTOP1_REG "\n\t"	/* optop[-4] = optop[-1] */\
	    "pushl %eax")		/* optop[-3] = eax */

  CODE(opc_dup2_x1, dup2_x1, ST0, ST2, 0) {
    __asm__("popl  %ecx\n\t"
	    "popl  %edx");	/* now state 2 */
    DUP2_X1_ST24(%ecx, %edx);
  }
  CODE(opc_dup2_x1, dup2_x1, ST1, ST4, 0) {
    __asm__("popl  %ecx");	/* now state 4 */
    DUP2_X1_ST24(%edx, %ecx);
  }
  CODE(opc_dup2_x1, dup2_x1, ST2, ST2, 0) {
    DUP2_X1_ST24(%ecx, %edx);
  }
  CODE(opc_dup2_x1, dup2_x1, ST3, ST2, 0) {
    __asm__("popl  %edx");	/* now state 2 */
    DUP2_X1_ST24(%ecx, %edx);
  }
  CODE(opc_dup2_x1, dup2_x1, ST4, ST4, 0) {
    DUP2_X1_ST24(%edx, %ecx);
  }

  /* dup2_x2 */
#define DUP2_X2_ST24(OPTOP1_REG, OPTOP2_REG) \
    __asm__("popl  %eax\n\t"		/* eax = optop[-3] */\
	    "popl  %edi\n\t"		/* edi = optop[-4] */\
	    "pushl " #OPTOP2_REG "\n\t"	/* optop[-4] = optop[-2] */\
	    "pushl " #OPTOP1_REG "\n\t"	/* optop[-3] = optop[-1] */\
	    "pushl %edi\n\t"		/* optop[-2] = edi */\
	    "pushl %eax");		/* optop[-1] = eax */

  CODE(opc_dup2_x2, dup2_x2, ST0, ST2, 0) {
    __asm__("popl  %ecx\n\t"
	    "popl  %edx");	/* now state 2 */
    DUP2_X2_ST24(%ecx, %edx);
  }
  CODE(opc_dup2_x2, dup2_x2, ST1, ST4, 0) {
    __asm__("popl  %ecx");	/* now state 4 */
    DUP2_X2_ST24(%edx, %ecx);
  }
  CODE(opc_dup2_x2, dup2_x2, ST2, ST2, 0) {
    DUP2_X2_ST24(%ecx, %edx);
  }
  CODE(opc_dup2_x2, dup2_x2, ST3, ST2, 0) {
    __asm__("popl  %edx");	/* now state 2 */
    DUP2_X2_ST24(%ecx, %edx);
  }
  CODE(opc_dup2_x2, dup2_x2, ST4, ST4, 0) {
    DUP2_X2_ST24(%edx, %ecx);
  }

  /* swap */
  CODE(opc_swap, swap, ST0, ST4, 0) {
    __asm__("popl  %ecx\n\t"
	    "popl  %edx");	/* now state 2 */
  }
  CODE(opc_swap, swap, ST1, ST2, 0) {
    __asm__("popl  %ecx");	/* now state 4 */
  }
  CODE(opc_swap, swap, ST2, ST4, 0) {}
  CODE(opc_swap, swap, ST3, ST4, 0) {
    __asm__("popl  %edx");	/* now state 2 */
  }
  CODE(opc_swap, swap, ST4, ST2, 0) {}

  /* iadd, isub, imul, iand, ior, ixor */
#ifdef RUNTIME_DEBUG
#  define ARITH_INT_DEBUG1(VAL1, VAL2, SYM) \
  if (runtime_debug) {\
    DEBUG_IN;\
    __asm__("pushl " #VAL2 "\n\t"\
	    "pushl " #VAL1);\
    PUSH_CONSTSTR("  %d " SYM " %d\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $12,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#else
#  define ARITH_INT_DEBUG1(VAL1, VAL2, SYM)
#endif

#define CODE_ARITH_INT(VOP, ROP, SYM) \
  CODE(opc_i##VOP, i##VOP, ST0, ST1, 0) {\
    __asm__("popl  %ecx\n\t"\
	    "popl  %edx");	/* now state 2 */\
    ARITH_INT_DEBUG1(%edx, %ecx, SYM);\
    __asm__(#ROP "l  %ecx,%edx");\
  }\
  CODE(opc_i##VOP, i##VOP, ST1, ST3, 0) {\
    __asm__("popl  %ecx");	/* now state 4 */\
    ARITH_INT_DEBUG1(%ecx, %edx, SYM);\
    __asm__(#ROP "l  %edx,%ecx");\
  }\
  CODE(opc_i##VOP, i##VOP, ST2, ST1, 0) {\
    ARITH_INT_DEBUG1(%edx, %ecx, SYM);\
    __asm__(#ROP "l  %ecx,%edx");\
  }\
  CODE(opc_i##VOP, i##VOP, ST3, ST1, 0) {\
    __asm__("popl  %edx");	/* now state 2 */\
    ARITH_INT_DEBUG1(%edx, %ecx, SYM);\
    __asm__(#ROP "l  %ecx,%edx");\
  }\
  CODE(opc_i##VOP, i##VOP, ST4, ST3, 0) {\
    ARITH_INT_DEBUG1(%ecx, %edx, SYM);\
    __asm__(#ROP "l  %edx,%ecx");\
  }

  CODE_ARITH_INT(add, add, "+");
  CODE_ARITH_INT(sub, sub, "-");
  CODE_ARITH_INT(mul, imul, "*");
  CODE_ARITH_INT(and, and, "&");
  CODE_ARITH_INT(or, or, "|");
  CODE_ARITH_INT(xor, xor, "^");

  /* idiv, irem */
#define INT_TEST(VOP, STATE, DIVIDER) \
    __asm__("testl " #DIVIDER "," #DIVIDER "\n\t"\
	    "jnz   i" #VOP "_st" #STATE "_1");\
    SIGNAL_ERROR0(ArithmeticException, STATE);\
    __asm__("i" #VOP "_st" #STATE "_1:")

#define CODE_ARITH_INT_TEST(VOP, RESULT_REG) \
  CODE(opc_i##VOP, i##VOP, ST0, ST3, 0) {\
    __asm__("popl  %ecx");	/* now state 3 */\
    INT_TEST(VOP, 0, %ecx);\
    __asm__("popl  %eax\n\t"\
	    "xorl  %edx,%edx\n\t"\
	    "cdq\n\t"\
	    "idivl %ecx\n\t"\
	    "movl  " #RESULT_REG ",%ecx");\
  }\
  CODE(opc_i##VOP, i##VOP, ST1, ST3, 0) {\
    INT_TEST(VOP, 1, %edx);\
    __asm__("popl  %eax\n\t"\
	    "movl  %edx,%ecx\n\t"\
	    "cdq\n\t"\
	    "idivl %ecx\n\t"\
	    "movl  " #RESULT_REG ",%ecx");\
  }\
  CODE(opc_i##VOP, i##VOP, ST2, ST3, 0) {\
    INT_TEST(VOP, 2, %ecx);\
    __asm__("movl  %edx,%eax\n\t"\
	    "cdq\n\t"\
	    "idivl %ecx\n\t"\
	    "movl  " #RESULT_REG ",%ecx");\
  }\
  CODE(opc_i##VOP, i##VOP, ST3, ST3, 0) {\
    INT_TEST(VOP, 3, %ecx);\
    __asm__("popl  %eax\n\t"\
	    "cdq\n\t"\
	    "idivl %ecx\n\t"	/* %eax ... %edx = %edx:%eax / %ecx */\
	    "movl  " #RESULT_REG ",%ecx");\
  }\
  CODE(opc_i##VOP, i##VOP, ST4, ST3, 0) {\
    INT_TEST(VOP, 4, %edx);\
    __asm__("movl  %ecx,%eax\n\t"\
	    "movl  %edx,%ecx\n\t"\
	    "cdq\n\t"\
	    "idivl %ecx\n\t"\
	    "movl  " #RESULT_REG ",%ecx");\
  }

  CODE_ARITH_INT_TEST(div, %eax);
  CODE_ARITH_INT_TEST(rem, %edx);

  /* ladd, lsub, land, lor, lxor */
#ifdef RUNTIME_DEBUG
#  define ARITH_LONG_DEBUG1(VAL1_LOW, VAL1_HIGH, VAL2_LOW, VAL2_HIGH, SYM) \
  if (runtime_debug) {\
    DEBUG_IN;\
    __asm__("pushl " #VAL2_HIGH "\n\tpushl " #VAL2_LOW "\n\t"\
	    "pushl " #VAL2_HIGH "\n\tpushl " #VAL2_LOW "\n\t"\
	    "pushl " #VAL1_HIGH "\n\tpushl " #VAL1_LOW "\n\t"\
	    "pushl " #VAL1_HIGH "\n\tpushl " #VAL1_LOW);\
    PUSH_CONSTSTR("  %lld(0x%016llx) " SYM " %lld(0x%016llx)\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $36,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#else
#  define ARITH_LONG_DEBUG1(VAL1_LOW, VAL1_HIGH, VAL2_LOW, VAL2_HIGH, SYM)
#endif

#define ARITH_LONG_SUB(ROP1, ROP2, OPTOP1_REG, OPTOP2_REG, SYM) \
    __asm__("popl  " #OPTOP1_REG "\n\t"	/* val1[0:31] */\
	    "popl  " #OPTOP2_REG);	/* val2[32:63] */\
    ARITH_LONG_DEBUG1(OPTOP1_REG, OPTOP2_REG, %eax, %edi, SYM);\
    __asm__(#ROP1 "l  %eax," #OPTOP1_REG "\n\t"	/* [0:31] */\
	    #ROP2 "l  %edi," #OPTOP2_REG)	/* [32:63] */

#define CODE_ARITH_LONG_SUB(VOP, ROP1, ROP2, SYM) \
  CODE(opc_l##VOP, l##VOP, ST0, ST2, 0) {\
    __asm__("popl  %eax\n\t"\
	    "popl  %edi");\
    ARITH_LONG_SUB(ROP1, ROP2, %ecx, %edx, SYM);\
  }\
  CODE(opc_l##VOP, l##VOP, ST1, ST4, 0) {\
    __asm__("movl  %edx,%eax\n\t"\
	    "popl  %edi");\
    ARITH_LONG_SUB(ROP1, ROP2, %edx, %ecx, SYM);\
  }\
  CODE(opc_l##VOP, l##VOP, ST2, ST2, 0) {\
    __asm__("movl  %ecx,%eax\n\t"\
	    "movl  %edx,%edi");\
    ARITH_LONG_SUB(ROP1, ROP2, %ecx, %edx, SYM);\
  }\
  CODE(opc_l##VOP, l##VOP, ST3, ST2, 0) {\
    __asm__("movl  %ecx,%eax\n\t"\
	    "popl  %edi");\
    ARITH_LONG_SUB(ROP1, ROP2, %ecx, %edx, SYM);\
  }\
  CODE(opc_l##VOP, l##VOP, ST4, ST4, 0) {\
    __asm__("movl  %edx,%eax\n\t"\
	    "movl  %ecx,%edi");\
    ARITH_LONG_SUB(ROP1, ROP2, %edx, %ecx, SYM);\
  }

#define ARITH_LONG(ROP1, ROP2, OPTOP1_REG, OPTOP2_REG, SYM) \
    __asm__("popl  %eax\n\t"	/* eax = val1[0:31] */\
	    "popl  %edi");	/* edi = val2[32:63] */\
    ARITH_LONG_DEBUG1(%eax, %edi, OPTOP1_REG, OPTOP2_REG, SYM);\
    __asm__(#ROP1 "l  %eax," #OPTOP1_REG "\n\t"	/* [0:31] */\
	    #ROP2 "l  %edi," #OPTOP2_REG)	/* [32:63] */

#define CODE_ARITH_LONG(VOP, ROP1, ROP2, SYM) \
  CODE(opc_l##VOP, l##VOP, ST0, ST2, 0) {\
    __asm__("popl  %ecx\n\t"\
	    "popl  %edx");		/* now state 2 */\
    ARITH_LONG(ROP1, ROP2, %ecx, %edx, SYM);\
  }\
  CODE(opc_l##VOP, l##VOP, ST1, ST4, 0) {\
    __asm__("popl  %ecx");		/* now state 4 */\
    ARITH_LONG(ROP1, ROP2, %edx, %ecx, SYM);\
  }\
  CODE(opc_l##VOP, l##VOP, ST2, ST2, 0) {\
    ARITH_LONG(ROP1, ROP2, %ecx, %edx, SYM);\
  }\
  CODE(opc_l##VOP, l##VOP, ST3, ST2, 0) {\
    __asm__("popl  %edx");		/* now state 2 */\
    ARITH_LONG(ROP1, ROP2, %ecx, %edx, SYM);\
  }\
  CODE(opc_l##VOP, l##VOP, ST4, ST4, 0) {\
    ARITH_LONG(ROP1, ROP2, %edx, %ecx, SYM);\
  }

  CODE_ARITH_LONG(add, add, adc, "+");
  CODE_ARITH_LONG_SUB(sub, sub, sbb, "-");
  CODE_ARITH_LONG(and, and, and, "&");
  CODE_ARITH_LONG(or, or, or, "|");
  CODE_ARITH_LONG(xor, xor, xor, "^");

  /* lmul */
#ifdef RUNTIME_DEBUG
#  define ARITH_LONG_CALL_DEBUG1(OPTOP1_REG, OPTOP2_REG, SYM) \
  if (runtime_debug) {\
    __asm__("pushl %eax\n\tpushl %edi");\
    __asm__("movl  8(%esp),%eax\n\t"\
	    "movl  12(%esp),%edi");\
    DEBUG_IN;\
    __asm__("pushl " #OPTOP2_REG "\n\t"\
	    "pushl " #OPTOP1_REG "\n\t"\
	    "pushl %edi\n\tpushl %eax");\
    PUSH_CONSTSTR("  %lld " SYM " %lld\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $20,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
    __asm__("popl %edi\n\tpopl %eax");\
  }
#else
#  define ARITH_LONG_CALL_DEBUG1(OPTOP1_REG, OPTOP2_REG, SYM)
#endif

#define LMUL_ST24(OPTOP1, LOW_WORD, HIGH_WORD) \
    __asm__("movl  " #OPTOP1 ",%eax\n\t"\
	    "movl  %edx,%edi\n\t"\
	    "mull  (%esp)\n\t"\
	    "movl  %eax,-8(%esp)\n\t"\
	    "movl  %edx,-4(%esp)\n\t"\
	    "imull 4(%esp)," #LOW_WORD "\n\t"\
	    "addl  " #LOW_WORD ",-4(%esp)\n\t"\
	    "imull (%esp)," #HIGH_WORD "\n\t"\
	    "addl  " #HIGH_WORD ",-4(%esp)\n\t"\
	    "movl  -8(%esp),%ecx\n\t"\
	    "movl  -4(%esp),%edx\n\t"	/* now state 2 */\
	    "addl  $8,%esp")

  CODE(opc_lmul, lmul, ST0, ST2, 0) {
    __asm__("popl  %ecx\n\t"
	    "popl  %edx");	/* now state 2 */
    ARITH_LONG_CALL_DEBUG1(%ecx, %edx, "*");
    LMUL_ST24(%ecx, %ecx, %edi);
  }
  CODE(opc_lmul, lmul, ST1, ST2, 0) {
    __asm__("popl  %ecx");	/* now state 4 */
    ARITH_LONG_CALL_DEBUG1(%edx, %ecx, "*");
    LMUL_ST24(%edx, %edi, %ecx);
  }
  CODE(opc_lmul, lmul, ST2, ST2, 0) {
    ARITH_LONG_CALL_DEBUG1(%ecx, %edx, "*");
    LMUL_ST24(%ecx, %ecx, %edi);
  }
  CODE(opc_lmul, lmul, ST3, ST2, 0) {
    __asm__("popl  %edx");	/* now state 2 */
    ARITH_LONG_CALL_DEBUG1(%ecx, %edx, "*");
    LMUL_ST24(%ecx, %ecx, %edi);
  }
  CODE(opc_lmul, lmul, ST4, ST2, 0) {
    ARITH_LONG_CALL_DEBUG1(%edx, %ecx, "*");
    LMUL_ST24(%edx, %edi, %ecx);
  }

  /* ldiv, lrem */
#define ARITH_LONG_CALL_ST24(ROP, OPTOP1_REG, OPTOP2_REG, SYM) \
    ARITH_LONG_CALL_DEBUG1(OPTOP1_REG, OPTOP2_REG, SYM);\
    __asm__("popl  %eax\n\t"	/* eax = v1[0:31] */\
	    "popl  %edi");	/* edi = v1[32:63] */\
    FUNCCALL_IN(0);\
    __asm__("pushl " #OPTOP2_REG "\n\t"	/* push v2[32:63] */\
	    "pushl " #OPTOP1_REG "\n\t"	/* push v2[0:31] */\
	    "pushl %edi\n\t"	/* push v1[32:63] */\
	    "pushl %eax\n\t"	/* push v1[0:31] */\
	    "call  " ROP "@PLT\n\t"\
	    "addl  $16,%esp");\
    __asm__(/* movl %edx,%edx */\
	    "movl  %eax,%ecx");\
    FUNCCALL_OUT(0)
	/* now state 2 */

#define LONG_TEST_ST24(VOP, STATE) \
	/* if ((%edx == 0) && (%ecx == 0))  throw Arith.Exc.; */\
    __asm__("movl  %edx,%eax\n\t"\
	    "orl   %ecx,%eax\n\t"\
	    "jnz   l" #VOP "_st" #STATE "_1");\
    SIGNAL_ERROR0(ArithmeticException, STATE);\
    __asm__("l" #VOP "_st" #STATE "_1:")
	/* now state [24] */

#define CODE_ARITH_LONG_TEST(VOP, ROP, SYM) \
  CODE(opc_l##VOP, l##VOP, ST0, ST2, 0) {\
    __asm__("popl  %ecx\n\t"\
	    "popl  %edx");	/* now state 2 */\
    LONG_TEST_ST24(VOP, 0);\
    ARITH_LONG_CALL_ST24(ROP, %ecx, %edx, SYM);\
  }\
  CODE(opc_l##VOP, l##VOP, ST1, ST2, 0) {\
    __asm__("popl  %ecx");	/* now state 4 */\
    LONG_TEST_ST24(VOP, 1);\
    ARITH_LONG_CALL_ST24(ROP, %edx, %ecx, SYM);\
  }\
  CODE(opc_l##VOP, l##VOP, ST2, ST2, 0) {\
    LONG_TEST_ST24(VOP, 2);\
    ARITH_LONG_CALL_ST24(ROP, %ecx, %edx, SYM);\
  }\
  CODE(opc_l##VOP, l##VOP, ST3, ST2, 0) {\
    __asm__("popl  %edx");	/* now state 2 */\
    LONG_TEST_ST24(VOP, 3);\
    ARITH_LONG_CALL_ST24(ROP, %ecx, %edx, SYM);\
  }\
  CODE(opc_l##VOP, l##VOP, ST4, ST2, 0) {\
    LONG_TEST_ST24(VOP, 4);\
    ARITH_LONG_CALL_ST24(ROP, %edx, %ecx, SYM);\
  }

  CODE_ARITH_LONG_TEST(div, SYMBOL(__divdi3), "/");
  CODE_ARITH_LONG_TEST(rem, SYMBOL(__moddi3), "mod");

  /* fadd, fsub, fmul, fdiv */
#ifdef RUNTIME_DEBUG
#  define ARITH_FLOAT_DEBUG1(SYM) \
  if (runtime_debug) {\
    __asm__("flds  4(%esp)\n\t"\
	    "flds  (%esp)");\
    DEBUG_IN;\
    __asm__("subl  $16,%esp\n\t"\
	    "fstpl 8(%esp)\n\t"\
	    "fstpl (%esp)");\
    PUSH_CONSTSTR("%g " SYM " %g = ");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $20,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#  define ARITH_FLOAT_DEBUG2 \
  if (runtime_debug) {\
    __asm__("flds  (%esp)");\
    DEBUG_IN;\
    __asm__("subl  $8,%esp\n\t"\
	    "fstpl (%esp)");\
    PUSH_CONSTSTR("%g\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $12,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#else
#  define ARITH_FLOAT_DEBUG1(SYM)
#  define ARITH_FLOAT_DEBUG2
#endif

#define CODE_ARITH_FLOAT(VOP, ROP, SYM) \
  CODE(opc_f##VOP, f##VOP, ST0, ST0, 0) {\
    ARITH_FLOAT_DEBUG1(SYM);\
    __asm__("flds  4(%esp)\n\t"\
	    "f" #ROP "s (%esp)\n\t"\
	    "addl  $4,%esp\n\t"\
	    "fstps (%esp)");\
    ARITH_FLOAT_DEBUG2;\
  }\
  CODE(opc_f##VOP, f##VOP, ST1, ST0, 0) {\
    __asm__("flds  (%esp)\n\t"\
	    "pushl %edx");\
    ARITH_FLOAT_DEBUG1(SYM);\
    __asm__("f" #ROP "s (%esp)\n\t"\
	    "addl  $4,%esp\n\t"\
	    "fstps (%esp)");\
    ARITH_FLOAT_DEBUG2;\
  }\
  CODE(opc_f##VOP, f##VOP, ST2, ST0, 0) {\
    __asm__("pushl %edx\n\t"\
	    "flds  (%esp)\n\t"\
	    "pushl %ecx");\
    ARITH_FLOAT_DEBUG1(SYM);\
    __asm__("f" #ROP "s (%esp)\n\t"\
	    "addl  $4,%esp\n\t"\
	    "fstps (%esp)");\
    ARITH_FLOAT_DEBUG2;\
  }\
  CODE(opc_f##VOP, f##VOP, ST3, ST0, 0) {\
    __asm__("flds  (%esp)\n\t"\
	    "pushl %ecx");\
    ARITH_FLOAT_DEBUG1(SYM);\
    __asm__("f" #ROP "s (%esp)\n\t"\
	    "addl  $4,%esp\n\t"\
	    "fstps (%esp)");\
    ARITH_FLOAT_DEBUG2;\
  }\
  CODE(opc_f##VOP, f##VOP, ST4, ST0, 0) {\
    __asm__("pushl %ecx\n\t"\
	    "flds  (%esp)\n\t"\
	    "pushl %edx\n\t"\
	    "f" #ROP "s (%esp)\n\t"\
	    "addl  $4,%esp\n\t"\
	    "fstps (%esp)");\
    ARITH_FLOAT_DEBUG2;\
  }

  CODE_ARITH_FLOAT(add, add, "+");
  CODE_ARITH_FLOAT(sub, sub, "-");
  CODE_ARITH_FLOAT(mul, mul, "*");
  CODE_ARITH_FLOAT(div, div, "/");

  /* dadd, dsub, dmul, ddiv */
#ifdef RUNTIME_DEBUG
#  define ARITH_DOUBLE_DEBUG1(OPTOP1_REG, OPTOP2_REG, SYM) \
  if (runtime_debug) {\
    __asm__("movl  (%esp),%edi\n\t"\
	    "movl  4(%esp),%eax");\
    DEBUG_IN;\
    __asm__("pushl " #OPTOP2_REG "\n\t"\
	    "pushl " #OPTOP1_REG "\n\t"\
	    "pushl %eax\n\tpushl %edi");\
    PUSH_CONSTSTR("  %g " SYM " %g = ");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $20,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#  define ARITH_DOUBLE_DEBUG2 \
  if (runtime_debug) {\
    __asm__("movl  (%esp),%edi\n\t"\
	    "movl  4(%esp),%eax");\
    DEBUG_IN;\
    __asm__("pushl %eax\n\tpushl %edi");\
    PUSH_CONSTSTR("%g\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $12,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#else
#  define ARITH_DOUBLE_DEBUG1(OPTOP1_REG, OPTOP2_REG, SYM)
#  define ARITH_DOUBLE_DEBUG2
#endif

#define DOUBLE_ST24(ROP, OPTOP1_REG, OPTOP2_REG, SYM) \
    ARITH_DOUBLE_DEBUG1(OPTOP1_REG, OPTOP2_REG, SYM);\
    __asm__("fldl  (%esp)\n\t"	/* load v1[0:63] */\
	    "addl  $8,%esp\n\t"\
	    "pushl " #OPTOP2_REG "\n\t"	/* v2[32:63] */\
	    "pushl " #OPTOP1_REG "\n\t"	/* v2[0:31] */\
	    "f" #ROP "l (%esp)\n\t"\
	    "fstpl (%esp)");\
    ARITH_DOUBLE_DEBUG2
	/* now state 0 */

#define CODE_ARITH_DOUBLE(VOP, ROP, SYM) \
  CODE(opc_d##VOP, d##VOP, ST0, ST0, 0) {\
    __asm__("popl  %ecx\n\t"\
	    "popl  %edx");		/* now state 2 */\
    DOUBLE_ST24(ROP, %ecx, %edx, SYM);\
  }\
  CODE(opc_d##VOP, d##VOP, ST1, ST0, 0) {\
    __asm__("popl  %ecx");		/* now state 4 */\
    DOUBLE_ST24(ROP, %edx, %ecx, SYM);\
  }\
  CODE(opc_d##VOP, d##VOP, ST2, ST0, 0) {\
    DOUBLE_ST24(ROP, %ecx, %edx, SYM);\
  }\
  CODE(opc_d##VOP, d##VOP, ST3, ST0, 0) {\
    __asm__("popl  %edx");		/* now state 2 */\
    DOUBLE_ST24(ROP, %ecx, %edx, SYM);\
  }\
  CODE(opc_d##VOP, d##VOP, ST4, ST0, 0) {\
    DOUBLE_ST24(ROP, %edx, %ecx, SYM);\
  }

  CODE_ARITH_DOUBLE(add, add, "+");
  CODE_ARITH_DOUBLE(sub, sub, "-");
  CODE_ARITH_DOUBLE(mul, mul, "*");
  CODE_ARITH_DOUBLE(div, div, "/");

  /* frem */
#define FREM_ST0 \
    ARITH_FLOAT_DEBUG1("mod");\
    __asm__("flds  (%esp)\n\t"		/* fld optop[-1].f (value2) */\
	    "flds  4(%esp)");		/* fld optop[-2].f (value1) */\
    FUNCCALL_IN(0);\
    __asm__("subl  $16,%esp");\
    __asm__("fstpl (%esp)\n\t"		/* stack top   = optop[-2] (value1) */\
	    "fstpl 8(%esp)\n\t"		/* stack top-1 = optop[-1] (value2) */\
	    "call  " SYMBOL(fmod) "@PLT\n\t"\
	    "addl  $16,%esp");\
    FUNCCALL_OUT(0);\
    __asm__("addl  $4,%esp\n\t"\
	    "fstps (%esp)");\
    ARITH_FLOAT_DEBUG2
	/* now state 0 */

  CODE(opc_frem, frem, ST0, ST0, 0) { FREM_ST0; }
  CODE(opc_frem, frem, ST1, ST0, 0) {
    __asm__("pushl %edx");	/* now state 0 */
    FREM_ST0;
  }
  CODE(opc_frem, frem, ST2, ST0, 0) {
    __asm__("pushl %edx\n\t"
	    "pushl %ecx");	/* now state 0 */
    FREM_ST0;
  }
  CODE(opc_frem, frem, ST3, ST0, 0) {
    __asm__("pushl %ecx");	/* now state 0 */
    FREM_ST0;
  }
  CODE(opc_frem, frem, ST4, ST0, 0) {
    __asm__("pushl %ecx\n\t"
	    "pushl %edx");	/* now state 0 */
    FREM_ST0;
  }

  /* drem */
#define DREM_ST24(OPTOP1_REG, OPTOP2_REG) \
    ARITH_DOUBLE_DEBUG1(OPTOP1_REG, OPTOP2_REG, "mod");\
    __asm__("popl  %eax\n\t"	/* eax = v1[0:31] */\
	    "popl  %edi");	/* edi = v1[32:63] */\
    FUNCCALL_IN(0);\
    __asm__("pushl " #OPTOP2_REG "\n\t"	/* push v2[32:63] */\
	    "pushl " #OPTOP1_REG "\n\t"	/* push v2[0:31] */\
	    "pushl %edi\n\t"	/* push v1[32:63] */\
	    "pushl %eax\n\t"	/* push v1[0:31] */\
	    "call  " SYMBOL(fmod) "@PLT\n\t"\
	    "addl  $16,%esp");\
    FUNCCALL_OUT(0);\
    __asm__("subl  $8,%esp\n\t"\
	    "fstpl (%esp)");\
    ARITH_DOUBLE_DEBUG2
	/* now state 0 */

  CODE(opc_drem, drem, ST0, ST0, 0) {
    __asm__("popl  %ecx\n\t"
	    "popl  %edx");	/* now state 2 */
    DREM_ST24(%ecx, %edx);
  }
  CODE(opc_drem, drem, ST1, ST0, 0) {
    __asm__("popl  %ecx");	/* now state 4 */
    DREM_ST24(%edx, %ecx);
  }
  CODE(opc_drem, drem, ST2, ST0, 0) {
    DREM_ST24(%ecx, %edx);
  }
  CODE(opc_drem, drem, ST3, ST0, 0) {
    __asm__("popl  %edx");	/* now state 2 */
    DREM_ST24(%ecx, %edx);
  }
  CODE(opc_drem, drem, ST4, ST0, 0) {
    DREM_ST24(%edx, %ecx);
  }

  /* ineg */
  CODE(opc_ineg, ineg, ST0, ST1, 0) {
    __asm__("popl  %edx\n\t"	/* now state 1 */
	    "negl  %edx");
    BIPUSH_DEBUG1(%edx);
  }
  CODE(opc_ineg, ineg, ST1, ST1, 0) {
    __asm__("negl  %edx");
    BIPUSH_DEBUG1(%edx);
  }
  CODE(opc_ineg, ineg, ST2, ST2, 0) {
    __asm__("negl  %ecx");
    BIPUSH_DEBUG1(%ecx);
  }
  CODE(opc_ineg, ineg, ST3, ST3, 0) {
    __asm__("negl  %ecx");
    BIPUSH_DEBUG1(%ecx);
  }
  CODE(opc_ineg, ineg, ST4, ST4, 0) {
    __asm__("negl  %edx");
    BIPUSH_DEBUG1(%edx);
  }

  /* lneg */
#define LNEG_ST24(OPTOP1_REG, OPTOP2_REG) \
    __asm__("negl  " #OPTOP1_REG "\n\t"\
	    "adcl  $0," #OPTOP2_REG "\n\t"\
	    "negl  " #OPTOP2_REG);\
    LDC2_W_DEBUG1(OPTOP1_REG, OPTOP2_REG)
	/* now state [24] */

  CODE(opc_lneg, lneg, ST0, ST2, 0) {
    __asm__("popl  %ecx\n\t"
	    "popl  %edx");	/* now state 2 */
    LNEG_ST24(%ecx, %edx);
  }
  CODE(opc_lneg, lneg, ST1, ST4, 0) {
    __asm__("popl  %ecx");	/* now state 4 */
    LNEG_ST24(%edx, %ecx);
  }
  CODE(opc_lneg, lneg, ST2, ST2, 0) {
    LNEG_ST24(%ecx, %edx);
  }
  CODE(opc_lneg, lneg, ST3, ST2, 0) {
    __asm__("popl  %edx");	/* now state 2 */
    LNEG_ST24(%ecx, %edx);
  }
  CODE(opc_lneg, lneg, ST4, ST4, 0) {
    LNEG_ST24(%edx, %ecx);
  }

  /* fneg */
#define FNEG_ST0 \
    __asm__("flds  (%esp)\n\t"\
	    "fchs\n\t"\
	    "fstps (%esp)")
	/* now state 0 */

  CODE(opc_fneg, fneg, ST0, ST0, 0) {
    FNEG_ST0;
  }
  CODE(opc_fneg, fneg, ST1, ST0, 0) {
    __asm__("pushl %edx");	/* now state 0 */
    FNEG_ST0;
  }
  CODE(opc_fneg, fneg, ST2, ST0, 0) {
    __asm__("pushl %edx\n\t"
	    "pushl %ecx");	/* now state 0 */
    FNEG_ST0;
  }
  CODE(opc_fneg, fneg, ST3, ST0, 0) {
    __asm__("pushl %ecx");	/* now state 0 */
    FNEG_ST0;
  }
  CODE(opc_fneg, fneg, ST4, ST0, 0) {
    __asm__("pushl %ecx\n\t"
	    "pushl %edx");	/* now state 0 */
    FNEG_ST0;
  }

  /* dneg */
#define DNEG_ST0 \
    __asm__("fldl  (%esp)\n\t"\
	    "fchs\n\t"\
	    "fstpl (%esp)")
	/* now state 0 */

  CODE(opc_dneg, dneg, ST0, ST0, 0) {
    DNEG_ST0;
  }
  CODE(opc_dneg, dneg, ST1, ST0, 0) {
    __asm__("pushl %edx");	/* now state 0 */
    DNEG_ST0;
  }
  CODE(opc_dneg, dneg, ST2, ST0, 0) {
    __asm__("pushl %edx\n\t"
	    "pushl %ecx");	/* now state 0 */
    DNEG_ST0;
  }
  CODE(opc_dneg, dneg, ST3, ST0, 0) {
    __asm__("pushl %ecx");	/* now state 0 */
    DNEG_ST0;
  }
  CODE(opc_dneg, dneg, ST4, ST0, 0) {
    __asm__("pushl %ecx\n\t"
	    "pushl %edx");	/* now state 0 */
    DNEG_ST0;
  }

  /* ishl */
#define SHIFT_INT_ST2(ROP) \
    __asm__(#ROP "l  %cl,%edx")
	/* now state 1 */

#define CODE_SHIFT_INT(VOP, ROP) \
  CODE(opc_i##VOP, i##VOP, ST0, ST1, 0) {\
    __asm__("popl  %ecx\n\t"\
	    "popl  %edx");	/* now state 2 */\
    SHIFT_INT_ST2(ROP);\
    BIPUSH_DEBUG1(%edx);\
  }\
  CODE(opc_i##VOP, i##VOP, ST1, ST1, 0) {\
    __asm__("movl  %edx,%ecx\n\t"\
	    "popl  %edx\n\t");	/* now state 2 */\
    SHIFT_INT_ST2(ROP);\
    BIPUSH_DEBUG1(%edx);\
  }\
  CODE(opc_i##VOP, i##VOP, ST2, ST1, 0) {\
    SHIFT_INT_ST2(ROP);\
    BIPUSH_DEBUG1(%edx);\
  }\
  CODE(opc_i##VOP, i##VOP, ST3, ST1, 0) {\
    __asm__("popl  %edx");	/* now state 2 */\
    SHIFT_INT_ST2(ROP);\
    BIPUSH_DEBUG1(%edx);\
  }\
  CODE(opc_i##VOP, i##VOP, ST4, ST1, 0) {\
    __asm__("xchg  %ecx,%edx");	/* now state 2 */\
    SHIFT_INT_ST2(ROP);\
    BIPUSH_DEBUG1(%edx);\
  }

  CODE_SHIFT_INT(shl, shl);
  CODE_SHIFT_INT(shr, sar);
  CODE_SHIFT_INT(ushr, shr);

  /* lshl, lshr, lushr */
	/*
	 * %cl is shift count.
	 * OP2 and OP3 are %eax and %edx.
	 */
#if 1	/* code generated by egcs-1.0.3 */
#define SHIFT_SIGNED_LONG(ROP64, ROP32, OP2_REG, OP3_REG, LABEL) \
    __asm__(#ROP64 "l %cl," #OP2_REG "," #OP3_REG "\n\t"\
	    #ROP32 "l %cl," #OP2_REG "\n\t"\
	    "testb $0x20,%cl\n\t"\
	    "jz    " LABEL "\n\t"\
	    "movl  " #OP2_REG "," #OP3_REG "\n\t"\
	    #ROP32 "l $0x1f," #OP2_REG "\n\t"\
	LABEL ":\n\t"\
	    "movl  %eax,%ecx")
	/* now state 2 or 4 */
#define SHIFT_UNSIGNED_LONG(ROP64, ROP32, OP2_REG, OP3_REG, LABEL) \
    __asm__(#ROP64 "l %cl," #OP2_REG "," #OP3_REG "\n\t"\
	    #ROP32 "l %cl," #OP2_REG "\n\t"\
	    "testb $0x20,%cl\n\t"\
	    "jz    " LABEL "\n\t"\
	    "movl  " #OP2_REG "," #OP3_REG "\n\t"\
	    "xorl  " #OP2_REG "," #OP2_REG "\n\t"	/* clear OP2_REG */\
	LABEL ":\n\t"\
	    "movl  %eax,%ecx")
	/* now state 2 or 4 */
#else	/* code generated by gcc-2.7.2.3 */
#define SHIFT_LONG(ROP64, ROP32, OP2_REG, OP3_REG, LABEL) \
    __asm__("rorb  %cl\n\t"\
	    #ROP64 "l %cl," #OP2_REG "," #OP3_REG "\n\t"\
	    #ROP32 "l  %cl," #OP2_REG "\n\t"\
	    #ROP64 "l %cl," #OP2_REG "," #OP3_REG "\n\t"\
	    #ROP32 "l  %cl," #OP2_REG "\n\t"\
	    "shrb  $7,%cl\n\t"\
	    #ROP64 "l %cl," #OP2_REG "," #OP3_REG "\n\t"\
	    #ROP32 "l  %cl," #OP2_REG "\n\t"\
	    "movl  %eax,%ecx")
	/* now state 2 or 4 */
#endif

#define CODE_SHIFT_LONG(SIGNEDP, VOP, ROP64, ROP32, REG_A, REG_B) \
  CODE(opc_l##VOP, l##VOP, ST0, ST2, 0) {\
    __asm__("popl  %ecx\n\t"		/* ecx = shift count */\
	    "popl  %eax\n\t"		/* eax = [0:31] */\
	    "popl  %edx");		/* edx = [32:63] */\
    SHIFT_##SIGNEDP##_LONG(ROP64, ROP32, REG_A, REG_B, #VOP "_st0");\
    LDC2_W_DEBUG1(%ecx, %edx);\
  }\
  CODE(opc_l##VOP, l##VOP, ST1, ST2, 0) {\
    __asm__("movl  %edx,%ecx\n\t"	/* ecx = shift count (%edx) */\
	    "popl  %eax\n\t"		/* eax = [0:31] */\
	    "popl  %edx");		/* edx = [32:63] */\
    SHIFT_##SIGNEDP##_LONG(ROP64, ROP32, REG_A, REG_B, #VOP "_st1");\
    LDC2_W_DEBUG1(%ecx, %edx);\
  }\
  CODE(opc_l##VOP, l##VOP, ST2, ST4, 0) {\
    __asm__("popl  %eax");		/* eax = [32:63] */\
    SHIFT_##SIGNEDP##_LONG(ROP64, ROP32, REG_B, REG_A, #VOP "_st2");\
    LDC2_W_DEBUG1(%edx, %ecx);\
  }\
  CODE(opc_l##VOP, l##VOP, ST3, ST2, 0) {\
    __asm__("popl  %eax\n\t"		/* eax = [0:31] */\
	    "popl  %edx");		/* edx = [32:63] */\
    SHIFT_##SIGNEDP##_LONG(ROP64, ROP32, REG_A, REG_B, #VOP "_st3");\
    LDC2_W_DEBUG1(%ecx, %edx);\
  }\
  CODE(opc_l##VOP, l##VOP, ST4, ST2, 0) {\
    __asm__("movl  %ecx,%eax\n\t"	/* eax = [0:31] (ecx) */\
	    "movl  %edx,%ecx\n\t"	/* ecx = shift count (edx) */\
	    "popl  %edx");		/* edx = [32:63] */\
    SHIFT_##SIGNEDP##_LONG(ROP64, ROP32, REG_A, REG_B, #VOP "_st4");\
    LDC2_W_DEBUG1(%ecx, %edx);\
  }

  CODE_SHIFT_LONG(UNSIGNED, shl, shld, shl, %eax, %edx);
  CODE_SHIFT_LONG(SIGNED, shr, shrd, sar, %edx, %eax);
  CODE_SHIFT_LONG(UNSIGNED, ushr, shrd, shr, %edx, %eax);

  /* iinc */
	/* const: (signed char *pc)[2], pc[1] * 4 */
#ifdef RUNTIME_DEBUG
#  define IINC_DEBUG1 \
  if (runtime_debug) {\
    DEBUG_IN;\
    __asm__("pushl " STR(CONST) "(%esi)\n\t"\
	    "pushl %eax\n\t"\
	    "movl  $" STR(CONST) ",%eax\n\t"\
	    "sarl  $2,%eax\n\t"\
	    "negl  %eax\n\t"\
	    "pushl %eax");\
    PUSH_CONSTSTR("  var[%d] + %d = %d\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $16,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#else
#  define IINC_DEBUG1
#endif

  CODE(opc_iinc, iinc, STANY, STSTA, 0) {
    __asm__("addl  $" STR(CONST) "," STR(CONST) "(%esi)");
    IINC_DEBUG1;
  }

  /* i2l */
  CODE(opc_i2l, i2l, ST0, ST2, 0) {
    __asm__("popl  %edx\n\t"	/* now state 1 */
	    "movl  %edx,%ecx\n\t"
	    "sarl  $31,%edx");
  }
  CODE(opc_i2l, i2l, ST1, ST2, 0) {
    __asm__("movl  %edx,%ecx\n\t"
	    "sarl  $31,%edx");
  }
  CODE(opc_i2l, i2l, ST2, ST4, 0) {
    __asm__("pushl %edx\n\t"	/* now state 3 */
	    "movl  %ecx,%edx\n\t"
	    "sarl  $31,%ecx");
  }
  CODE(opc_i2l, i2l, ST3, ST4, 0) {
    __asm__("movl  %ecx,%edx\n\t"
	    "sarl  $31,%ecx");
  }
  CODE(opc_i2l, i2l, ST4, ST2, 0) {
    __asm__("pushl %ecx\n\t"	/* now state 1 */
	    "movl  %edx,%ecx\n\t"
	    "sarl  $31,%edx");
  }

  /* i2f */
#define I2F_ST0 \
    __asm__("fildl (%esp)\n\t"\
	    "fstps (%esp)")

  CODE(opc_i2f, i2f, ST0, ST0, 0) {
    I2F_ST0;
  }
  CODE(opc_i2f, i2f, ST1, ST0, 0) {
    __asm__("pushl %edx");	/* now state 0 */
    I2F_ST0;
  }
  CODE(opc_i2f, i2f, ST2, ST0, 0) {
    __asm__("pushl %edx\n\t"
	    "pushl %ecx");	/* now state 0 */
    I2F_ST0;
  }
  CODE(opc_i2f, i2f, ST3, ST0, 0) {
    __asm__("pushl %ecx");	/* now state 0 */
    I2F_ST0;
  }
  CODE(opc_i2f, i2f, ST4, ST0, 0) {
    __asm__("pushl %ecx\n\t"
	    "pushl %edx");	/* now state 0 */
    I2F_ST0;
  }

  /* i2d */
#define I2D_ST0 \
    __asm__("fildl (%esp)\n\t"\
	    "fstpl -4(%esp)\n\t"\
	    "subl  $4,%esp")

  CODE(opc_i2d, i2d, ST0, ST0, 0) {
    I2D_ST0;
  }
  CODE(opc_i2d, i2d, ST1, ST0, 0) {
    __asm__("pushl %edx");	/* now state 0 */
    I2D_ST0;
  }
  CODE(opc_i2d, i2d, ST2, ST0, 0) {
    __asm__("pushl %edx\n\t"
	    "pushl %ecx");	/* now state 0 */
    I2D_ST0;
  }
  CODE(opc_i2d, i2d, ST3, ST0, 0) {
    __asm__("pushl %ecx");	/* now state 0 */
    I2D_ST0;
  }
  CODE(opc_i2d, i2d, ST4, ST0, 0) {
    __asm__("pushl %ecx\n\t"
	    "pushl %edx");	/* now state 0 */
    I2D_ST0;
  }

  /* l2i */
  CODE(opc_l2i, l2i, ST0, ST1, 0) {
    __asm__("popl  %edx\n\t"	/* now state 1 */
	    "addl  $4,%esp");
  }
  CODE(opc_l2i, l2i, ST1, ST1, 0) {
    __asm__("addl  $4,%esp");
  }
  CODE(opc_l2i, l2i, ST2, ST3, 0) {}
  CODE(opc_l2i, l2i, ST3, ST3, 0) {
    __asm__("addl  $4,%esp");
  }
  CODE(opc_l2i, l2i, ST4, ST1, 0) {}

  /* l2f */
#define L2F_ST0 \
    __asm__("fildll (%esp)\n\t"\
	    "fstps 4(%esp)\n\t"\
	    "addl  $4,%esp")

  CODE(opc_l2f, l2f, ST0, ST0, 0) {
    L2F_ST0;
  }
  CODE(opc_l2f, l2f, ST1, ST0, 0) {
    __asm__("pushl %edx");	/* now state 0 */
    L2F_ST0;
  }
  CODE(opc_l2f, l2f, ST2, ST0, 0) {
    __asm__("pushl %edx\n\t"
	    "pushl %ecx");	/* now state 0 */
    L2F_ST0;
  }
  CODE(opc_l2f, l2f, ST3, ST0, 0) {
    __asm__("pushl %ecx");	/* now state 0 */
    L2F_ST0;
  }
  CODE(opc_l2f, l2f, ST4, ST0, 0) {
    __asm__("pushl %ecx\n\t"
	    "pushl %edx");	/* now state 0 */
    L2F_ST0;
  }

  /* l2d */
#define L2D_ST0 \
    __asm__("fildll (%esp)\n\t"\
	    "fstpl (%esp)")

  CODE(opc_l2d, l2d, ST0, ST0, 0) {
    L2D_ST0;
  }
  CODE(opc_l2d, l2d, ST1, ST0, 0) {
    __asm__("pushl %edx");	/* now state 0 */
    L2D_ST0;
  }
  CODE(opc_l2d, l2d, ST2, ST0, 0) {
    __asm__("pushl %edx\n\t"
	    "pushl %ecx");	/* now state 0 */
    L2D_ST0;
  }
  CODE(opc_l2d, l2d, ST3, ST0, 0) {
    __asm__("pushl %ecx");	/* now state 0 */
    L2D_ST0;
  }
  CODE(opc_l2d, l2d, ST4, ST0, 0) {
    __asm__("pushl %ecx\n\t"
	    "pushl %edx");	/* now state 0 */
    L2D_ST0;
  }

  /* f2i, f2l, d2i, d2l */
#ifdef RUNTIME_DEBUG
#  define REAL2INT_DEBUG1 \
  if (runtime_debug) {\
    DEBUG_IN;\
    __asm__("subl  $8,%esp\n\t"\
	    "fstl  (%esp)");\
    PUSH_CONSTSTR("  %g\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $12,%esp");\
    DEBUG_OUT;\
  }
#else
#  define REAL2INT_DEBUG1
#endif

#define REAL2INT_ST0(FLD, ADJUST_STACK, MINUS_ADJ, FIST) \
    __asm__(#FLD "  (%esp)");\
    REAL2INT_DEBUG1;\
    __asm__("fstcw -6(%esp)\n\t"	/* original control word */\
	    "ftst\n\t"\
	    "movw  -6(%esp),%ax\n\t"\
	    "orw   $(3<<10),%ax\n\t"	/* rounding mode */\
	    "movw  %ax,-8(%esp)\n\t"	/* modified control word */\
	    "fldcw -8(%esp)\n\t"\
	    ADJUST_STACK "\n\t"\
	    #FIST " (%esp)\n\t"\
	    "fldcw (" #MINUS_ADJ "-6)(%esp)")
	/* now state 0 */

#define REAL2INT_F2I_ST0	REAL2INT_ST0(flds, "", 0, fistpl)
#define REAL2INT_F2L_ST0	REAL2INT_ST0(flds, "subl $4,%esp", 4, fistpll)
#define REAL2INT_D2I_ST0	REAL2INT_ST0(fldl, "addl $4,%esp", -4, fistpl)
#define REAL2INT_D2L_ST0	REAL2INT_ST0(fldl, "", 0, fistpll)

#define REAL2INT_CHECK_INT(VOP, STATE) \
    __asm__("popl  %edx\n\t"	/* now state 1 */\
	    "cmpl  $0x80000000,%edx\n\t"\
	    "jne   " #VOP "_st" #STATE "_done\n\t"\
	    "fnstsw %ax\n\t"\
	    "sahf\n\t"\
	    "jp    " #VOP "_st" #STATE "_nan\n\t"	/* jump if not NaN */\
	    "adcl  $-1,%edx\n\t"	/* carry flag is set if result < 0 */\
	    "jmp   " #VOP "_st" #STATE "_done\n\t"\
	#VOP "_st" #STATE "_nan:\n\t"	/* result is NaN: return 0 */\
	    "xorl  %edx,%edx\n\t"\
	#VOP "_st" #STATE "_done:")
	/* now state 1 */

#define REAL2INT_CHECK_LONG(VOP, STATE) \
    __asm__("popl  %ecx\n\t"	/* ecx = low word */\
	    "popl  %edx\n\t"	/* edx = high word */\
		/* now state 2 */\
	/* check if high word is 0x80000000 and low word is 0 */\
	    "movl  %edx,%eax\n\t"\
	    "xorl  $0x80000000,%eax\n\t"\
	    "orl   %ecx,%eax\n\t"\
	    "jnz   " #VOP "_st" #STATE "_done\n\t"\
	    "fnstsw %ax\n\t"\
	    "sahf\n\t"\
	    "jp    " #VOP "_st" #STATE "_nan\n\t"	/* jump if not NaN */\
	    "adcl  $-1,%ecx\n\t"	/* carry flag is set if result < 0 */\
	    "adcl  $-1,%edx\n\t"\
	    "jmp   " #VOP "_st" #STATE "_done\n\t"\
	#VOP "_st" #STATE "_nan:\n\t"	/* result is NaN: return 0 */\
	    "xorl  %edx,%edx\n\t"\
	    "xorl  %ecx,%ecx\n\t"\
	#VOP "_st" #STATE "_done:")
	/* now state 2 */

#define CODE_REAL2INT(vop, VOP, RET_TYPE, LAST_STATE) \
  CODE(opc_##vop, vop, ST0, ST##LAST_STATE, 0) {\
    REAL2INT_##VOP##_ST0;\
    REAL2INT_CHECK_##RET_TYPE(vop, 0);\
  }\
  CODE(opc_##vop, vop, ST1, ST##LAST_STATE, 0) {\
    __asm__("pushl %edx");	/* now state 0 */\
    REAL2INT_##VOP##_ST0;\
    REAL2INT_CHECK_##RET_TYPE(vop, 1);\
  }\
  CODE(opc_##vop, vop, ST2, ST##LAST_STATE, 0) {\
    __asm__("pushl %edx\n\t"\
	    "pushl %ecx");	/* now state 0 */\
    REAL2INT_##VOP##_ST0;\
    REAL2INT_CHECK_##RET_TYPE(vop, 2);\
  }\
  CODE(opc_##vop, vop, ST3, ST##LAST_STATE, 0) {\
    __asm__("pushl %ecx");	/* now state 0 */\
    REAL2INT_##VOP##_ST0;\
    REAL2INT_CHECK_##RET_TYPE(vop, 3);\
  }\
  CODE(opc_##vop, vop, ST4, ST##LAST_STATE, 0) {\
    __asm__("pushl %ecx\n\t"\
	    "pushl %edx");	/* now state 0 */\
    REAL2INT_##VOP##_ST0;\
    REAL2INT_CHECK_##RET_TYPE(vop, 4);\
  }

  CODE_REAL2INT(f2i, F2I, INT, 1);
  CODE_REAL2INT(f2l, F2L, LONG, 2);
  CODE_REAL2INT(d2i, D2I, INT, 1);
  CODE_REAL2INT(d2l, D2L, LONG, 2);

  /* f2d */
#define F2D_ST0 \
    __asm__("flds  (%esp)\n\t"\
	    "fstpl -4(%esp)\n\t"\
	    "subl  $4,%esp")

  CODE(opc_f2d, f2d, ST0, ST0, 0) {
    F2D_ST0;
  }
  CODE(opc_f2d, f2d, ST1, ST0, 0) {
    __asm__("pushl %edx");	/* now state 0 */\
    F2D_ST0;
  }
  CODE(opc_f2d, f2d, ST2, ST0, 0) {
    __asm__("pushl %edx\n\t"\
	    "pushl %ecx");	/* now state 0 */\
    F2D_ST0;
  }
  CODE(opc_f2d, f2d, ST3, ST0, 0) {
    __asm__("pushl %ecx");	/* now state 0 */\
    F2D_ST0;
  }
  CODE(opc_f2d, f2d, ST4, ST0, 0) {
    __asm__("pushl %ecx\n\t"\
	    "pushl %edx");	/* now state 0 */\
    F2D_ST0;
  }

  /* d2f */
#define D2F_ST0 \
    __asm__("fldl  (%esp)\n\t"\
	    "fstps 4(%esp)\n\t"\
	    "addl  $4,%esp")

  CODE(opc_d2f, d2f, ST0, ST0, 0) {
    D2F_ST0;
  }
  CODE(opc_d2f, d2f, ST1, ST0, 0) {
    __asm__("pushl %edx");	/* now state 0 */\
    D2F_ST0;
  }
  CODE(opc_d2f, d2f, ST2, ST0, 0) {
    __asm__("pushl %edx\n\t"\
	    "pushl %ecx");	/* now state 0 */\
    D2F_ST0;
  }
  CODE(opc_d2f, d2f, ST3, ST0, 0) {
    __asm__("pushl %ecx");	/* now state 0 */\
    D2F_ST0;
  }
  CODE(opc_d2f, d2f, ST4, ST0, 0) {
    __asm__("pushl %ecx\n\t"\
	    "pushl %edx");	/* now state 0 */\
    D2F_ST0;
  }

  /* i2b, i2c, i2s */
#define I2B(REG) \
    __asm__("shl  $24," #REG "\n\t"\
	    "sar  $24," #REG)
#define I2C(REG) \
    __asm__("shl  $16," #REG "\n\t"\
	    "shr  $16," #REG)
#define I2S(REG) \
    __asm__("shl  $16," #REG "\n\t"\
	    "sar  $16," #REG)

#define CODE_I2BCS(vop, VOP) \
  CODE(opc_##vop, vop, ST0, ST1, 0) {\
    __asm__("popl  %edx");	/* now state 1 */\
    VOP(%edx);\
  }\
  CODE(opc_##vop, vop, ST1, ST1, 0) {\
    VOP(%edx);\
  }\
  CODE(opc_##vop, vop, ST2, ST2, 0) {\
    VOP(%ecx);\
  }\
  CODE(opc_##vop, vop, ST3, ST3, 0) {\
    VOP(%ecx);\
  }\
  CODE(opc_##vop, vop, ST4, ST4, 0) {\
    VOP(%edx);\
  }

  CODE_I2BCS(i2b, I2B);
  CODE_I2BCS(i2c, I2C);
  CODE_I2BCS(i2s, I2S);

  /* lcmp */
#define LCMP_ST24(OPTOP1_REG, OPTOP2_REG, STATE) \
    __asm__("popl  %eax\n\t"	/* eax = v1[0:31] */\
	    "popl  %edi");	/* edi = v1[32:63] */\
    ARITH_LONG_DEBUG1(%eax, %edi, OPTOP1_REG, OPTOP2_REG, "");\
    __asm__("cmpl  " #OPTOP2_REG ",%edi\n\t"	/* cmp v1[32:63] - v2[32:63]*/\
	    "jl    lcmp_st" #STATE "_lt\n\t"\
	    "jg    lcmp_st" #STATE "_ge\n\t"\
	    "cmpl  " #OPTOP1_REG ",%eax\n\t"	/* cmp v1[0:31] - v2[0:31] */\
	    "jb    lcmp_st" #STATE "_lt\n\t"\
	"lcmp_st" #STATE "_ge:\n\t"\
	    "movl  $0,%edx\n\t"\
	    "setnz %dl\n\t"\
	    "jmp   lcmp_st" #STATE "_done\n\t"\
	"lcmp_st" #STATE "_lt:\n\t"\
	    "movl  $-1,%edx\n\t"\
	"lcmp_st" #STATE "_done:")
	/* now state 1 */

  CODE(opc_lcmp, lcmp, ST0, ST1, 0) {
    __asm__("popl  %ecx\n\t"
	    "popl  %edx");	/* now state 2 */
    LCMP_ST24(%ecx, %edx, 0);
  }
  CODE(opc_lcmp, lcmp, ST1, ST1, 0) {
    __asm__("popl  %ecx");	/* now state 4 */
    LCMP_ST24(%edx, %ecx, 1);
  }
  CODE(opc_lcmp, lcmp, ST2, ST1, 0) {
    LCMP_ST24(%ecx, %edx, 2);
  }
  CODE(opc_lcmp, lcmp, ST3, ST1, 0) {
    __asm__("popl  %edx");	/* now state 2 */
    LCMP_ST24(%ecx, %edx, 3);
  }
  CODE(opc_lcmp, lcmp, ST4, ST1, 0) {
    LCMP_ST24(%edx, %ecx, 4);
  }

  /* fcmpl, fcmpg, dcmpl, dcmpg */
#ifdef RUNTIME_DEBUG
#  define FCMP_DEBUG1 \
  if (runtime_debug) {\
    __asm__("flds  4(%esp)\n\t"\
	    "flds  (%esp)");\
    DEBUG_IN;\
    __asm__("subl  $16,%esp\n\t"\
	    "fstpl 8(%esp)\n\t"\
	    "fstpl (%esp)");\
    PUSH_CONSTSTR("  %g %g\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $20,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#  define DCMP_DEBUG1 \
  if (runtime_debug) {\
    __asm__("movl  (%esp),%edi\n\t"\
	    "movl  4(%esp),%eax\n\t"\
	    "movl  8(%esp),%ecx\n\t"\
	    "movl  12(%esp),%edx");\
    DEBUG_IN;\
    __asm__("pushl %eax\n\tpushl %edi\n\t"\
	    "pushl %edx\n\tpushl %ecx");\
    PUSH_CONSTSTR("  %g %g\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $20,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#else
#  define FCMP_DEBUG1
#  define DCMP_DEBUG1
#endif

#define FCMPL_COMPARE_ST0 \
    FCMP_DEBUG1;\
    __asm__("flds  4(%esp)\n\t"\
	    "fcomps (%esp)\n\t"\
	    "addl  $8,%esp")
#define FCMPG_COMPARE_ST0	FCMPL_COMPARE_ST0

#define DCMPL_COMPARE_ST0 \
    DCMP_DEBUG1;\
    __asm__("fldl  8(%esp)\n\t"\
	    "fcompl (%esp)\n\t"\
	    "addl  $16,%esp")
#define DCMPG_COMPARE_ST0	DCMPL_COMPARE_ST0

#define FLOAT_CMP_NAN_g(LABEL) \
	    "jnp  " LABEL "_normal\n\t"/* PF indicates NaN */\
	    "movl  $1,%edx\n\t"\
	    "jmp  " LABEL "_done\n\t"
#define FLOAT_CMP_NAN_l(LABEL) \
	    "jp   " LABEL "_l\n\t"	/* PF indicates NaN */

#define FLOAT_CMP_ST0(SUF, LABEL) \
    __asm__("fnstsw %ax\n\t"\
	    "sahf\n\t"\
	    FLOAT_CMP_NAN_##SUF(LABEL)\
	LABEL "_normal:\n\t"\
	    "jc    " LABEL "_l\n\t"\
	LABEL "_g:\n\t"\
	    "movl  $0,%edx\n\t"\
	    "setnz %dl\n\t"\
	    "jmp   " LABEL "_done\n\t"\
	LABEL "_l:\n\t"\
	    "movl  $-1,%edx\n\t"\
	LABEL "_done:")
	/* now state 1 */

#define CODE_FLOAT_CMP(vop, VOP, SUF) \
  CODE(opc_##vop, vop, ST0, ST1, 0) {\
    VOP##_COMPARE_ST0;\
    FLOAT_CMP_ST0(SUF, #VOP "_st0");\
  }\
  CODE(opc_##vop, vop, ST1, ST1, 0) {\
    __asm__("pushl %edx");	/* now state 0 */\
    VOP##_COMPARE_ST0;\
    FLOAT_CMP_ST0(SUF, #VOP "_st1");\
  }\
  CODE(opc_##vop, vop, ST2, ST1, 0) {\
    __asm__("pushl %edx\n\t"\
	    "pushl %ecx");	/* now state 0 */\
    VOP##_COMPARE_ST0;\
    FLOAT_CMP_ST0(SUF, #VOP "_st2");\
  }\
  CODE(opc_##vop, vop, ST3, ST1, 0) {\
    __asm__("pushl %ecx");	/* now state 0 */\
    VOP##_COMPARE_ST0;\
    FLOAT_CMP_ST0(SUF, #VOP "_st3");\
  }\
  CODE(opc_##vop, vop, ST4, ST1, 0) {\
    __asm__("pushl %ecx\n\t"\
	    "pushl %edx");	/* now state 0 */\
    VOP##_COMPARE_ST0;\
    FLOAT_CMP_ST0(SUF, #VOP "_st4");\
  }

  CODE_FLOAT_CMP(fcmpl, FCMPL, l);
  CODE_FLOAT_CMP(fcmpg, FCMPG, g);
  CODE_FLOAT_CMP(dcmpl, DCMPL, l);
  CODE_FLOAT_CMP(dcmpg, DCMPG, g);

  /* ifeq, ifne, iflt, ifge, ifgt, ifle */
#ifdef RUNTIME_DEBUG
#  define IF_DEBUG1(OPTOP1_REG) \
  if (runtime_debug) {\
    DEBUG_IN;\
    __asm__("pushl " #OPTOP1_REG);\
    PUSH_CONSTSTR("  %d\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $8,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#else
#  define IF_DEBUG1(OPTOP1_REG)
#endif

#define IF(OPTOP1_REG, JP_ROP) \
    IF_DEBUG1(OPTOP1_REG);\
    __asm__("testl " #OPTOP1_REG "," #OPTOP1_REG)
    /* jump: #JP_ROP STR(ADDR_JP) */

#define CODE_IF(VOP, JP_ROP) \
  CODE(opc_if##VOP, if##VOP, ST0, ST0, 0) {\
    __asm__("popl  %eax");\
    IF(%eax, JP_ROP);\
  }\
  CODE(opc_if##VOP, if##VOP, ST1, ST0, 0) {\
    IF(%edx, JP_ROP);\
  }\
  CODE(opc_if##VOP, if##VOP, ST2, ST1, 0) {\
    IF(%ecx, JP_ROP);\
  }\
  CODE(opc_if##VOP, if##VOP, ST3, ST0, 0) {\
    IF(%ecx, JP_ROP);\
  }\
  CODE(opc_if##VOP, if##VOP, ST4, ST3, 0) {\
    IF(%edx, JP_ROP);\
  }

  CODE_IF(eq, je);
  CODE_IF(ne, jne);
  CODE_IF(lt, jl);
  CODE_IF(ge, jge);
  CODE_IF(gt, jg);
  CODE_IF(le, jle);

  /* if_icmp{eq,ne,lt,ge,gt,le} */
#ifdef RUNTIME_DEBUG
#  define IF_ICMP_DEBUG(OPTOP1_REG, OPTOP2_REG) \
  if (runtime_debug) {\
    DEBUG_IN;\
    __asm__("pushl " #OPTOP1_REG "\n\t"\
	    "pushl " #OPTOP2_REG);\
    PUSH_CONSTSTR("  %d %d\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $12,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#else
#  define IF_ICMP_DEBUG(OPTOP1_REG, OPTOP2_REG)
#endif

#define IF_ICMP_ST24(OPTOP1_REG, OPTOP2_REG, JP_ROP) \
    IF_ICMP_DEBUG(OPTOP1_REG, OPTOP2_REG);\
    __asm__("cmpl  " #OPTOP1_REG "," #OPTOP2_REG)
    /* jump: #JP_ROP STR(ADDR_JP) */
	/* now state 0 */

#define CODE_IF_ICMP(VOP, JP_ROP) \
  CODE(opc_if_icmp##VOP, if_icmp##VOP, ST0, ST0, 0) {\
    __asm__("popl  %ecx\n\t"\
	    "popl  %edx");	/* now state 2 */\
    IF_ICMP_ST24(%ecx, %edx, JP_ROP);\
  }\
  CODE(opc_if_icmp##VOP, if_icmp##VOP, ST1, ST0, 0) {\
    __asm__("popl  %ecx");	/* now state 4 */\
    IF_ICMP_ST24(%edx, %ecx, JP_ROP);\
  }\
  CODE(opc_if_icmp##VOP, if_icmp##VOP, ST2, ST0, 0) {\
    IF_ICMP_ST24(%ecx, %edx, JP_ROP);\
  }\
  CODE(opc_if_icmp##VOP, if_icmp##VOP, ST3, ST0, 0) {\
    __asm__("popl  %edx");	/* now state 2 */\
    IF_ICMP_ST24(%ecx, %edx, JP_ROP);\
  }\
  CODE(opc_if_icmp##VOP, if_icmp##VOP, ST4, ST0, 0) {\
    IF_ICMP_ST24(%edx, %ecx, JP_ROP);\
  }

  CODE_IF_ICMP(eq, je);
  CODE_IF_ICMP(ne, jne);
  CODE_IF_ICMP(lt, jl);
  CODE_IF_ICMP(ge, jge);
  CODE_IF_ICMP(gt, jg);
  CODE_IF_ICMP(le, jle);

  /* goto */
  CODE(opc_goto, goto, STANY, STSTA, 0) {
    /* jump: "jmp   " STR(ADDR_JP) */
  }

  /* jsr */
	/* const: native offset of a next instruction */
#ifdef RUNTIME_DEBUG
#  define JSR_DEBUG1(REG) \
  if (runtime_debug) {\
    DEBUG_IN;\
    __asm__("pushl " #REG);\
    PUSH_CONSTSTR("  push 0x%08x\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $8,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#else
#  define JSR_DEBUG1(REG)
#endif

#define JSR(REG) \
    __asm__("movl  $" STR(CONST) "," #REG);\
    JSR_DEBUG1(REG)
    /* jump: "jmp  " STR(ADDR_JP) */

  CODE(opc_jsr, jsr, ST0, ST1, 0) {
    JSR(%edx);
  }
  CODE(opc_jsr, jsr, ST1, ST2, 0) {
    JSR(%ecx);
  }
  CODE(opc_jsr, jsr, ST2, ST4, 0) {
    __asm__("pushl %edx");	/* now state 3 */
    JSR(%edx);
  }
  CODE(opc_jsr, jsr, ST3, ST4, 0) {
    JSR(%edx);
  }
  CODE(opc_jsr, jsr, ST4, ST2, 0) {
    __asm__("pushl %ecx");	/* now state 1 */
    JSR(%ecx);
  }

  /* ret */
	/* const: index * 4 */
  CODE(opc_ret, ret, STANY, STSTA, 0) {
    __asm__("movl " STR(CONST) "(%esi),%eax");	/* eax = vars[index] */
    ILOAD_DEBUG1(%eax);
    COMPILEDCODE(%edi);	/* edi = mb->CompiledCode */
    __asm__("addl  %edi,%eax");
    /* jump: "jmp   *%eax" */
  }

  /* tableswitch */
	/* const: low, high, table offset (in native) */
  /*
   * table:  Each element is 8 byte.
   *  An element consists of an offset of target in native code
   *  and a pointer to trampoline code.
   *	default, offset(low), offset(low+1), ..., offset(high)
   */
#ifdef RUNTIME_DEBUG
#  define TBLSW_DEBUG1(INDEX) \
  if (runtime_debug) {\
    DEBUG_IN;\
    __asm__("pushl %eax\n\tpushl %edi\n\tpushl " #INDEX);\
    PUSH_CONSTSTR("  index: %d (%d - %d)\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $16,%esp");\
    DEBUG_OUT;\
  }
#  define TBLSW_DEBUG2(OFF) \
  if (runtime_debug) {\
    DEBUG_IN;\
    __asm__("pushl " #OFF "\n\tpushl " #OFF);\
    PUSH_CONSTSTR("  native off: 0x%08x(%d)\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $12,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#else
#  define TBLSW_DEBUG1(INDEX)
#  define TBLSW_DEBUG2(OFF)
#endif

#define TBLSW(INDEX, STATE) \
    __asm__("movl  $" STR(CONST) ",%edi\n\t"	/* low */\
	    "movl  $" STR(CONST) ",%eax");	/* high */\
    TBLSW_DEBUG1(INDEX);\
    __asm__("subl  %edi," #INDEX "\n\t"		/* index -= low */\
	    "subl  %edi,%eax");			/* high -= low */\
    \
    COMPILEDCODE(%edi);				/* edi = mb->CompiledCode */\
    __asm__("pushl %edi");	/* push mb->CompiledCode */\
    \
    __asm__("addl  $" STR(CONST) ",%edi"); /* edi += offset of the table */\
		/* edi = addr. of the table */\
    __asm__("cmpl  " #INDEX ",%eax\n\t"		/* test high - index */\
	    "jb    tblsw_st" #STATE "_default\n\t"\
	    "leal  8(%edi," #INDEX ",8),%edi\n\t"	/* edi = tgt offset */\
	"tblsw_st" #STATE "_default:\n\t"\
	    "movl  (%edi),%eax");\
    TBLSW_DEBUG2(%eax);\
    \
    __asm__("addl  (%esp),%eax\n\t"	/* eax += mb->CompiledCode */\
	    "addl  $4,%esp");\
		/* eax = target address */\
    \
    __asm__("jmp   *4(%edi)")

  CODE(opc_tableswitch, tableswitch, ST0, ST0, 0) {
    __asm__("popl  %edx");	/* now state 1 */
    TBLSW(%edx, 0);
  }
  CODE(opc_tableswitch, tableswitch, ST1, ST0, 0) {
    TBLSW(%edx, 1);
  }
  CODE(opc_tableswitch, tableswitch, ST2, ST1, 0) {
    TBLSW(%ecx, 2);
  }
  CODE(opc_tableswitch, tableswitch, ST3, ST0, 0) {
    TBLSW(%ecx, 3);
  }
  CODE(opc_tableswitch, tableswitch, ST4, ST3, 0) {
    TBLSW(%edx, 4);
  }

  /* lookupswitch */
	/* const: npairs, table offset (in native) */
  /*
   * table:  Each element is 12 byte.
   *  An element consists of a key, an offset of target in native code
   *  and a pointer to trampoline code.
   *
   *	element(1), element(2), ..., element(npairs), element(default)
   */
#ifdef RUNTIME_DEBUG
#  define LUSW_DEBUG1(KEY) \
  if (runtime_debug) {\
    DEBUG_IN;\
    __asm__("pushl %eax\n\tpushl " #KEY);\
    PUSH_CONSTSTR("  key: %d, npairs: %d\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $12,%esp");\
    DEBUG_OUT;\
  }
#  define LUSW_DEBUG2 \
  if (runtime_debug) {\
    DEBUG_IN;\
    __asm__("pushl 8(%edi)\n\t"\
	    "pushl 4(%edi)\n\tpushl 4(%edi)\n\t"\
	    "pushl (%edi)");\
    PUSH_CONSTSTR("  match: %d, target offset: 0x%08x(%d), trampoline code: 0x%08x\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $20,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#else
#  define LUSW_DEBUG1(KEY)
#  define LUSW_DEBUG2
#endif

#define LUSW(KEY, STATE) \
    COMPILEDCODE(%edi);\
    __asm__("pushl %edi");	/* push mb->CompiledCode */\
    \
    __asm__("movl  $" STR(CONST) ",%eax");	/* npairs */\
    __asm__("addl  $" STR(CONST) ",%edi");	/* offset of the table */\
		/* edi = addr. of the table */\
    LUSW_DEBUG1(KEY);\
    __asm__(\
	"lusw_st" #STATE "_loop:\n\t"\
	    "cmpl  (%edi)," #KEY "\n\t"\
	    "je    lusw_st" #STATE "_loopend\n\t"\
	    "addl  $12,%edi\n\t"\
	    "decl  %eax\n\t"\
	    "jnz   lusw_st" #STATE "_loop\n\t"\
	"lusw_st" #STATE "_loopend:");\
    LUSW_DEBUG2;\
    __asm__("movl  4(%edi),%eax");\
    \
    __asm__("addl  (%esp),%eax\n\t"	/* eax += mb->CompiledCode */\
	    "addl  $4,%esp");\
		/* eax = tgt addr. */\
    \
    __asm__("jmp   *8(%edi)")


  CODE(opc_lookupswitch, lookupswitch, ST0, ST0, 0) {
    __asm__("popl  %edx");	/* now state 1 */
    LUSW(%edx, 0);
  }
  CODE(opc_lookupswitch, lookupswitch, ST1, ST0, 0) {
    LUSW(%edx, 1);
  }
  CODE(opc_lookupswitch, lookupswitch, ST2, ST1, 0) {
    LUSW(%ecx, 2);
  }
  CODE(opc_lookupswitch, lookupswitch, ST3, ST0, 0) {
    LUSW(%ecx, 3);
  }
  CODE(opc_lookupswitch, lookupswitch, ST4, ST3, 0) {
    LUSW(%edx, 4);
  }

  /* ireturn */
#define IRETURN(OPTOP1_REG) \
    /* (ee->current_frame->prev->optop++)[0] = OPTOP1_REG */\
    {\
      register ExecEnv *cur_ee asm("eax");\
      cur_ee = ee;\
    }\
    __asm__("movl  " EE_CURRENTFRAME(%eax) ",%edi\n\t"\
	    "movl  " FRAME_PREV(%edi) ",%eax\n\t"\
	    "movl  " FRAME_OPTOP(%eax) ",%edi\n\t"\
	    "movl  " #OPTOP1_REG ",(%edi)\n\t"\
	    "addl  $4," FRAME_OPTOP(%eax));\
    BIPUSH_DEBUG1(OPTOP1_REG);\
    __asm__("jmp   " STR(ADDR_FIN))

  CODE(opc_ireturn, [ifa]return, ST0, ST0, 0) {
    __asm__("popl  %edx");	/* now state 1 */
    IRETURN(%edx);
  }
  CODE(opc_ireturn, [ifa]return, ST1, ST0, 0) {
    IRETURN(%edx);
  }
  CODE(opc_ireturn, [ifa]return, ST2, ST1, 0) {
    IRETURN(%ecx);
  }
  CODE(opc_ireturn, [ifa]return, ST3, ST0, 0) {
    IRETURN(%ecx);
  }
  CODE(opc_ireturn, [ifa]return, ST4, ST3, 0) {
    IRETURN(%edx);
  }

  /* lreturn */
#define LRETURN(OPTOP1_REG, OPTOP2_REG) \
    /* (ee->current_frame->prev->optop++)[0] = OPTOP1_REG\
       (ee->current_frame->prev->optop++)[1] = OPTOP2_REG */\
    {\
      register ExecEnv *cur_ee asm("eax");\
      cur_ee = ee;\
    }\
    __asm__("movl  " EE_CURRENTFRAME(%eax) ",%edi\n\t"\
	    "movl  " FRAME_PREV(%edi) ",%eax\n\t"\
	    "movl  " FRAME_OPTOP(%eax) ",%edi\n\t"\
	    "movl  " #OPTOP1_REG ",(%edi)\n\t"\
	    "movl  " #OPTOP2_REG ",4(%edi)\n\t"\
	    "addl  $8," FRAME_OPTOP(%eax));\
    LDC2_W_DEBUG1(OPTOP1_REG, OPTOP2_REG);\
    __asm__("jmp   " STR(ADDR_FIN))

  CODE(opc_lreturn, [ld]return, ST0, ST0, 0) {
    __asm__("popl  %ecx\n\t"
	    "popl  %edx");	/* now state 2 */
    LRETURN(%ecx, %edx);
  }
  CODE(opc_lreturn, [ld]return, ST1, ST0, 0) {
    __asm__("popl  %ecx");	/* now state 4 */
    LRETURN(%edx, %ecx);
  }
  CODE(opc_lreturn, [ld]return, ST2, ST0, 0) {
    LRETURN(%ecx, %edx);
  }
  CODE(opc_lreturn, [ld]return, ST3, ST0, 0) {
    __asm__("popl  %edx");	/* now state 2 */
    LRETURN(%ecx, %edx);
  }
  CODE(opc_lreturn, [ld]return, ST4, ST0, 0) {
    LRETURN(%edx, %ecx);
  }

  /* return */
  CODE(opc_return, return, STANY, STSTA, 0) {
    __asm__("jmp   " STR(ADDR_FIN));
  }

  /* getstatic */
	/* const: address */
  CODE(opc_getstatic, getstatic, ST0, ST1, 0) {
    __asm__("movl  (" STR(CONST) "),%edx");
    BIPUSH_DEBUG1(%edx);
  }
  CODE(opc_getstatic, getstatic, ST1, ST2, 0) {
    __asm__("movl  (" STR(CONST) "),%ecx");
    BIPUSH_DEBUG1(%ecx);
  }
  CODE(opc_getstatic, getstatic, ST2, ST4, 0) {
    __asm__("pushl %edx\n\t"	/* now state 3 */
	    "movl  (" STR(CONST) "),%edx");
    BIPUSH_DEBUG1(%edx);
  }
  CODE(opc_getstatic, getstatic, ST3, ST4, 0) {
    __asm__("movl  (" STR(CONST) "),%edx");
    BIPUSH_DEBUG1(%edx);
  }
  CODE(opc_getstatic, getstatic, ST4, ST2, 0) {
    __asm__("pushl %ecx\n\t"	/* now state 1 */
	    "movl  (" STR(CONST) "),%ecx");
    BIPUSH_DEBUG1(%ecx);
  }

	/* const: address, address + 4 */
#define GETSTATIC2_ST0(LOW_REG, HIGH_REG) \
    __asm__("movl  (" STR(CONST) ")," #LOW_REG "\n\t"\
	    "movl  (" STR(CONST) ")," #HIGH_REG)
  CODE(opc_getstatic2, getstatic2, ST0, ST2, 0) {
    GETSTATIC2_ST0(%ecx, %edx);
  }
  CODE(opc_getstatic2, getstatic2, ST1, ST2, 0) {
    __asm__("pushl %edx");	/* now state 0 */
    GETSTATIC2_ST0(%ecx, %edx);
  }
  CODE(opc_getstatic2, getstatic2, ST2, ST4, 0) {
    __asm__("pushl %edx\n\t"\
	    "pushl %ecx");	/* now state 0 */\
    GETSTATIC2_ST0(%edx, %ecx);
  }
  CODE(opc_getstatic2, getstatic2, ST3, ST4, 0) {
    __asm__("pushl %ecx");	/* now state 0 */\
    GETSTATIC2_ST0(%edx, %ecx);
  }
  CODE(opc_getstatic2, getstatic2, ST4, ST2, 0) {
    __asm__("pushl %ecx\n\t"\
	    "pushl %edx");	/* now state 0 */\
    GETSTATIC2_ST0(%ecx, %edx);
  }

  /* putstatic */
	/* const: address */
  CODE(opc_putstatic, putstatic, ST0, ST0, 0) {
    __asm__("popl  %edx\n\t"	/* now state 1 */
	    "movl  %edx,(" STR(CONST) ")");
    BIPUSH_DEBUG1(%edx);
  }
  CODE(opc_putstatic, putstatic, ST1, ST0, 0) {
    __asm__("movl  %edx,(" STR(CONST) ")");
    BIPUSH_DEBUG1(%edx);
  }
  CODE(opc_putstatic, putstatic, ST2, ST1, 0) {
    __asm__("movl  %ecx,(" STR(CONST) ")");
    BIPUSH_DEBUG1(%ecx);
  }
  CODE(opc_putstatic, putstatic, ST3, ST0, 0) {
    __asm__("movl  %ecx,(" STR(CONST) ")");
    BIPUSH_DEBUG1(%ecx);
  }
  CODE(opc_putstatic, putstatic, ST4, ST3, 0) {
    __asm__("movl  %edx,(" STR(CONST) ")");
    BIPUSH_DEBUG1(%edx);
  }

	/* const: address, address + 4 */
#define PUTSTATIC2(OPTOP1_REG, OPTOP2_REG) \
    __asm__("movl  " #OPTOP1_REG ",(" STR(CONST) ")\n\t"\
	    "movl  " #OPTOP2_REG ",(" STR(CONST) ")")
	/* now state 0 */
  
  CODE(opc_putstatic2, putstatic2, ST0, ST0, 0) {
    __asm__("popl  %ecx\n\t"
	    "popl  %edx");	/* now state 2 */
    PUTSTATIC2(%ecx, %edx);
  }
  CODE(opc_putstatic2, putstatic2, ST1, ST0, 0) {
    __asm__("popl  %ecx");	/* now state 4 */
    PUTSTATIC2(%edx, %ecx);
  }
  CODE(opc_putstatic2, putstatic2, ST2, ST0, 0) {
    PUTSTATIC2(%ecx, %edx);
  }
  CODE(opc_putstatic2, putstatic2, ST3, ST0, 0) {
    __asm__("popl  %edx");	/* now state 2 */
    PUTSTATIC2(%ecx, %edx);
  }
  CODE(opc_putstatic2, putstatic2, ST4, ST0, 0) {
    PUTSTATIC2(%edx, %ecx);
  }

  /* getfield */
	/* const: slot */
#ifndef NO_CHECK
#  define FIELD_ACC(HANDLE, VOP, STATE) \
    __asm__("movl  $" STR(CONST) ",%eax");\
	/* slot: fb->u.offset / sizeof(OBJECT) */\
    NULL_TEST(HANDLE, #VOP "_st" #STATE "_1", STATE)
#else
#  define FIELD_ACC(HANDLE, VOP, STATE) \
    __asm__("movl  $" STR(CONST) ",%eax")
	/* fb->u.offset / sizeof(OBJECT) */
#endif	/* NO_CHECK */
	/* eax = index */

#ifdef JITDO
#  define JITDO_GETFIELD(HANDLE, TGT, LABEL) \
    JUMP_IF_NOT_REMOTE(LABEL "_local");\
    JUMP_IF_NOT_PROXY(HANDLE, LABEL "_local");\
    \
    __asm__("pushl %eax\n\t"	/* slot */\
	    "pushl " #HANDLE);	/* obj (Proxy) */\
    __asm__("pushl %0" : : "m" (ee));	/* ee */\
    \
    __asm__("movl  $" STR(CONST) ",%edi");\
    __asm__("testl %edi,%edi\n\t"\
	    "jnz   " LABEL "_obj\n\t"\
	    "call  " SYMBOL(proxy_get32field) "@PLT\n\t"\
	    "jmp   " LABEL "_remoteget_done\n\t"\
	    LABEL "_obj:\n\t"\
	    "call  " SYMBOL(proxy_getobjfield) "@PLT\n\t"\
	    LABEL "_remoteget_done:\n\t"\
); DEBUG_IN;\
__asm__("pushl %eax");\
PUSH_CONSTSTR("JITDO_GETFIELD: 0x%08x\n");\
__asm__("call  printf@PLT\n\t"\
	"addl  $8,%esp");\
FFLUSH;\
DEBUG_OUT; __asm__(\
	    "addl  $12,%esp\n\t"\
	    "movl  %eax," #TGT "\n\t"\
	    "jmp  " LABEL "_done\n\t"\
	    LABEL "_local:")
#else
#  define JITDO_GETFIELD(HANDLE, TGT, LABEL)
#endif	/* JITDO */

  CODE(opc_getfield, getfield, ST0, ST3, 1) {
    __asm__("popl  %edx");	/* now state 1 */
    FIELD_ACC(%edx, getfield, 0);
    JITDO_GETFIELD(%edx, %ecx, "getfield_st0");
    OBJ_GETSLOT(%edx, %eax, %ecx);
    __asm__("getfield_st0_done:");
    ILOAD_DEBUG1(%ecx);
  }
  CODE(opc_getfield, getfield, ST1, ST3, 1) {
    FIELD_ACC(%edx, getfield, 1);
    JITDO_GETFIELD(%edx, %ecx, "getfield_st1");
    OBJ_GETSLOT(%edx, %eax, %ecx);
    __asm__("getfield_st1_done:");
    ILOAD_DEBUG1(%ecx);
  }
  CODE(opc_getfield, getfield, ST2, ST2, 1) {
    FIELD_ACC(%ecx, getfield, 2);
    JITDO_GETFIELD(%ecx, %ecx, "getfield_st2");
    OBJ_GETSLOT(%ecx, %eax, %ecx);
    __asm__("getfield_st2_done:");
    ILOAD_DEBUG1(%ecx);
  }
  CODE(opc_getfield, getfield, ST3, ST1, 1) {
    FIELD_ACC(%ecx, getfield, 3);
    JITDO_GETFIELD(%ecx, %edx, "getfield_st3");
    OBJ_GETSLOT(%ecx, %eax, %edx);
    __asm__("getfield_st3_done:");
    ILOAD_DEBUG1(%edx);
  }
  CODE(opc_getfield, getfield, ST4, ST4, 1) {
    FIELD_ACC(%edx, getfield, 4);
    JITDO_GETFIELD(%edx, %edx, "getfield_st4");
    OBJ_GETSLOT(%edx, %eax, %edx);
    __asm__("getfield_st4_done:");
    ILOAD_DEBUG1(%edx);
  }

#ifdef JITDO
#  define JITDO_GETFIELD2(HANDLE, TGT_LOW, TGT_HIGH, LABEL) \
    JUMP_IF_NOT_REMOTE(LABEL "_local");\
    JUMP_IF_NOT_PROXY(HANDLE, LABEL "_local");\
    \
    __asm__("pushl %eax\n\t"	/* slot */\
	    "pushl " #HANDLE);	/* obj (Proxy) */\
    __asm__("pushl %0" : : "m" (ee));	/* ee */\
    \
    __asm__("call  " SYMBOL(proxy_get64field) "@PLT\n\t"\
	    "movl  %edx," #TGT_HIGH "\n\t"\
	    "addl  $12,%esp\n\t"\
	    "movl  %eax," #TGT_LOW "\n\t"\
	    "jmp  " LABEL "_done\n\t"\
	    LABEL "_local:")
#else
#  define JITDO_GETFIELD2(HANDLE, TGT_LOW, TGT_HIGH, LABEL)
#endif	/* JITDO */

	/* const: slot */
  CODE(opc_getfield2, getfield2, ST0, ST2, 1) {
    __asm__("popl  %edx");	/* now state 1 */
    FIELD_ACC(%edx, getfield2, 0);
    JITDO_GETFIELD2(%edx, %ecx, %edx, "getfield2_st0");
    OBJ_GETSLOT2(%edx, %eax, %ecx, %edx);
    __asm__("getfield2_st0_done:");
    LLOAD_DEBUG1(%ecx, %edx);
  }
  CODE(opc_getfield2, getfield2, ST1, ST2, 1) {
    FIELD_ACC(%edx, getfield2, 1);
    JITDO_GETFIELD2(%edx, %ecx, %edx, "getfield2_st1");
    OBJ_GETSLOT2(%edx, %eax, %ecx, %edx);
    __asm__("getfield2_st1_done:");
    LLOAD_DEBUG1(%ecx, %edx);
  }
  CODE(opc_getfield2, getfield2, ST2, ST4, 1) {
    __asm__("pushl %edx");	/* now state 3 */
    FIELD_ACC(%ecx, getfield2, 2);
    JITDO_GETFIELD2(%ecx, %edx, %ecx, "getfield2_st2");
    OBJ_GETSLOT2(%ecx, %eax, %edx, %ecx);
    __asm__("getfield2_st2_done:");
    LLOAD_DEBUG1(%edx, %ecx);
  }
  CODE(opc_getfield2, getfield2, ST3, ST4, 1) {
    FIELD_ACC(%ecx, getfield2, 3);
    JITDO_GETFIELD2(%ecx, %edx, %ecx, "getfield2_st3");
    OBJ_GETSLOT2(%ecx, %eax, %edx, %ecx);
    __asm__("getfield2_st3_done:");
    LLOAD_DEBUG1(%edx, %ecx);
  }
  CODE(opc_getfield2, getfield2, ST4, ST2, 1) {
    __asm__("pushl  %ecx");	/* now state 1 */
    FIELD_ACC(%edx, getfield2, 4);
    JITDO_GETFIELD2(%edx, %ecx, %edx, "getfield2_st4");
    OBJ_GETSLOT2(%edx, %eax, %ecx, %edx);
    __asm__("getfield2_st4_done:");
    LLOAD_DEBUG1(%ecx, %edx);
  }

  /* putfield */
	/* const: slot */
#ifdef JITDO
#  define JITDO_PUTFIELD(HANDLE, VAL, LABEL) \
    JUMP_IF_NOT_REMOTE(LABEL "_local");\
    JUMP_IF_NOT_PROXY(HANDLE, LABEL "_local");\
    \
    __asm__("pushl " #VAL "\n\t"/* value */\
	    "pushl %eax\n\t"	/* slot */\
	    "pushl " #HANDLE);	/* obj (Proxy) */\
    __asm__("pushl %0" : : "m" (ee));	/* ee */\
    \
    __asm__("movl  $" STR(CONST) ",%edi");\
    __asm__("testl %edi,%edi\n\t"\
	    "jnz   " LABEL "_obj\n\t"\
	    "call  " SYMBOL(proxy_put32field) "@PLT\n\t"\
	    "addl  $16,%esp\n\t"\
	    "jmp   " LABEL "_done\n\t"\
	    LABEL "_obj:\n\t"\
	    "call  " SYMBOL(proxy_putobjfield) "@PLT\n\t"\
	    "addl  $16,%esp\n\t"\
	    "jmp   " LABEL "_done\n\t"\
	    LABEL "_local:")
#else
#  define JITDO_PUTFIELD(HANDLE, VAL, LABEL)
#endif	/* JITDO */

#define PUTFIELD_ST24(OPTOP1_REG, OPTOP2_REG, STATE) \
    FIELD_ACC(OPTOP2_REG, putfield, STATE);\
    JITDO_PUTFIELD(OPTOP2_REG, OPTOP1_REG, "putfield_st" #STATE);\
    OBJ_SETSLOT(OPTOP2_REG, %eax, OPTOP1_REG);\
    __asm__("putfield_st" #STATE "_done:")

  CODE(opc_putfield, putfield, ST0, ST0, 1) {
    __asm__("popl  %ecx\n\t"
	    "popl  %edx");	/* now state 2 */
    PUTFIELD_ST24(%ecx, %edx, 0);
    ILOAD_DEBUG1(%ecx);
  }
  CODE(opc_putfield, putfield, ST1, ST0, 1) {
    __asm__("popl  %ecx");	/* now state 4 */
    PUTFIELD_ST24(%edx, %ecx, 1);
    ILOAD_DEBUG1(%edx);
  }
  CODE(opc_putfield, putfield, ST2, ST0, 1) {
    PUTFIELD_ST24(%ecx, %edx, 2);
    ILOAD_DEBUG1(%ecx);
  }
  CODE(opc_putfield, putfield, ST3, ST0, 1) {
    __asm__("popl  %edx");	/* now state 2 */
    PUTFIELD_ST24(%ecx, %edx, 3);
    ILOAD_DEBUG1(%ecx);
  }
  CODE(opc_putfield, putfield, ST4, ST0, 1) {
    PUTFIELD_ST24(%edx, %ecx, 4);
    ILOAD_DEBUG1(%edx);
  }

#ifdef JITDO
#  define JITDO_PUTFIELD2(HANDLE, VAL_LOW, VAL_HIGH, LABEL) \
    __asm__("pushl %eax\n\tpushl %edi");	/* save */\
    JUMP_IF_NOT_REMOTE(LABEL "_local");\
    __asm__("movl  (%esp),%eax");	/* %eax = TOS(%edi) */\
    JUMP_IF_NOT_PROXY(%eax, LABEL "_local");	/* 1st arg must not be %edi */\
    __asm__("popl  %edi\n\tpopl  %eax");	/* restore */\
    \
    __asm__("pushl " #VAL_HIGH "\n\t"/* value */\
	    "pushl " #VAL_LOW "\n\t"/* value */\
	    "pushl %eax\n\t"	/* slot */\
	    "pushl " #HANDLE);	/* obj (Proxy) */\
    __asm__("pushl %0" : : "m" (ee));	/* ee */\
    \
    __asm__("call  " SYMBOL(proxy_put64field) "@PLT\n\t"\
	    "addl  $20,%esp\n\t"\
	    "jmp   " LABEL "_done\n\t"\
	    LABEL "_local:\n\t"\
	    "popl  %edi\n\tpopl  %eax")		/* restore */
#else
#  define JITDO_PUTFIELD2(HANDLE, VAL_LOW, VAL_HIGH, LABEL)
#endif	/* JITDO */

	/* const: slot */
#define PUTFIELD2_ST24(OPTOP1_REG, OPTOP2_REG, STATE) \
    __asm__("popl  %edi");	/* edi = handle */\
    FIELD_ACC(%edi, putfield2, STATE);\
    JITDO_PUTFIELD2(%edi, OPTOP1_REG, OPTOP2_REG, "putfield2_st" #STATE);\
    OBJ_SETSLOT2(%edi, %eax, OPTOP1_REG, OPTOP2_REG);\
    __asm__("putfild2_st" #STATE "_done:")

  CODE(opc_putfield2, putfield2, ST0, ST0, 1) {
    __asm__("popl  %ecx\n\t"
	    "popl  %edx");	/* now state 2 */
    PUTFIELD2_ST24(%ecx, %edx, 0);
    LLOAD_DEBUG1(%ecx, %edx);
  }
  CODE(opc_putfield2, putfield2, ST1, ST0, 1) {
    __asm__("popl  %ecx");	/* now state 4 */
    PUTFIELD2_ST24(%edx, %ecx, 1);
    LLOAD_DEBUG1(%edx, %ecx);
  }
  CODE(opc_putfield2, putfield2, ST2, ST0, 1) {
    PUTFIELD2_ST24(%ecx, %edx, 2);
    LLOAD_DEBUG1(%ecx, %edx);
  }
  CODE(opc_putfield2, putfield2, ST3, ST0, 1) {
    __asm__("popl  %edx");	/* now state 2 */
    PUTFIELD2_ST24(%ecx, %edx, 3);
    LLOAD_DEBUG1(%ecx, %edx);
  }
  CODE(opc_putfield2, putfield2, ST4, ST0, 1) {
    PUTFIELD2_ST24(%edx, %ecx, 4);
    LLOAD_DEBUG1(%edx, %ecx);
  }

  /* invokevirtual */
	/* const: args_size, methodIndex, retsize */
  /* invokespecial */
	/* const: args_size, method, local_var_space, retsize */
  /* invokestatic */
	/* const: args_size, method, local_var_space, retsize */
  /* invokeinterface */
	/* const: args_size, guessptr, imethod, retsize */
#ifdef RUNTIME_DEBUG
#  define CALL_INVOKEMETHOD_DEBUG1 \
  if (runtime_debug) {\
    DEBUG_IN;\
    __asm__("pushl %eax");\
    PUSH_CONSTSTR("  invokeMethod() returns: %d\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $8,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#else
#  define CALL_INVOKEMETHOD_DEBUG1
#endif

#ifdef RUNTIME_DEBUG
#  define CALL_INVOKEMETHOD_FUNC \
  __asm__("pushl %0" : : "m" (runtime_debug));\
  __asm__("call  " SYMBOL(invokeMethod) "@PLT\n\t"\
	  "addl  $32,%esp")
#else
#  define CALL_INVOKEMETHOD_FUNC \
  __asm__("call  " SYMBOL(invokeMethod) "@PLT\n\t"\
	  "addl  $28,%esp")
#endif

#define CALL_INVOKEMETHOD(LABEL) \
    __asm__("pushl $" STR(CONST) "\n\t"	/* retsize */\
	    "pushl %edi");		/* args_size */\
    __asm__("pushl %0" : : "m" (ee));	/* ee */\
    CALL_INVOKEMETHOD_FUNC;\
    CALL_INVOKEMETHOD_DEBUG1;\
    \
    __asm__("popl  %edi\n\tpopl  %esi");	/* restore */\
    \
    __asm__("popl  %ecx");		/* free local var space */\
    __asm__("addl  %ecx,%esp");\
    \
    INVOKE_RESTACK(LABEL)

#define INVOKE_RESTACK(LABEL) \
	/* restack\
	   eax: return value from invokeMethod()\
	   edi: arg_size */\
    \
    __asm__(LABEL "_restack:");\
	/* adjust optop */\
    __asm__("leal  (%esp,%edi,4),%esp");/* esp += (args_size * 4) */\
	/* if (eax < 0)  goto exception_handler */\
    __asm__("testl %eax,%eax\n\t"\
	    "jl    " STR(ADDR_EXC));\
    \
    __asm__("jz    " LABEL "_done");	/* if (ret_size == 0) done */\
    \
    /* edi = ee->current_frame->optop */\
    {\
      register ExecEnv *cur_ee asm("edi");\
      cur_ee = ee;\
    }\
    __asm__("movl  " EE_CURRENTFRAME(%edi) ",%ecx\n\t"\
	    "movl  " FRAME_OPTOP(%ecx) ",%edi");\
    \
    __asm__("cmpl  $2,%eax\n\t"\
	    "jne   " LABEL "_1");	/* if (ret_size != 2) jump */\
		/* ret_size: 2 */\
    __asm__("subl  $8," FRAME_OPTOP(%ecx) "\n\t"\
	    "pushl -4(%edi)\n\t"\
	    "pushl -8(%edi)\n\t"\
	    "jmp   " LABEL "_done\n\t"\
	    LABEL "_1:");\
		/* ret_size: 1 */\
    __asm__("subl  $4," FRAME_OPTOP(%ecx) "\n\t"\
	    "pushl -4(%edi)\n\t"\
	    LABEL "_done:")
		/* now state 0 */


#ifdef JITDO
#  define JITDO_INVOKE(LABEL) \
    __asm__("movl  %esp,%ecx");	/* save %esp to %ecx */\
    \
    __asm__("pushl %edi");	/* save */\
    JUMP_IF_NOT_REMOTE(LABEL "_local");\
    JUMP_IF_NOT_PROXY(%edx /* is obj */, LABEL "_local");\
    \
    /* call Proxy#* at local */\
    __asm__("pushl %eax");	/* save */\
    METHOD_CLAZZ(%eax,%eax);\
    CB_METHODTABLE(%eax,%eax);\
    __asm__("movl  " SYMBOL(proxy_methodtable) "@GOT(%ebx),%edi");\
    __asm__("movl  (%edi),%edi");\
DEBUG_IN;\
__asm__("pushl %edi\n\tpushl %eax");\
PUSH_CONSTSTR("eax: 0x%08x, edi: 0x%08x\n");\
__asm__("call printf@PLT\n\t"\
	"addl $12,%esp");\
FFLUSH;\
DEBUG_OUT;\
    __asm__("cmpl  %eax,%edi\n\t"\
	    "popl  %eax\n\t"	/* restore */\
	    "jz    " LABEL "_local");\
    \
    __asm__("pushl %esi");\
    \
    __asm__("pushl %ecx\n\t"	/* stack pointer */\
	    "pushl " METHOD_SIGNATURE(%eax) "\n\t"\
	    "pushl " METHOD_NAME(%eax) "\n\t"\
	    "pushl %edx");	/* obj (Proxy) */\
    __asm__("pushl %0" : : "m" (ee));	/* ee */\
    __asm__("call  " SYMBOL(proxy_invoke) "@PLT\n\t"\
	    "addl  $20,%esp");\
    \
    __asm__("popl  %esi\n\tpopl  %edi");\
	/* %eax: ret_size, %edi: args_size */\
    \
    __asm__("jmp   " LABEL "_restack\n\t"\
	    LABEL "_local:");\
    __asm__("popl  %edi")	/* restore */
#else
#  define JITDO_INVOKE(LABEL)
#endif	/* JITDO */

#define INVOKE_ST0(VOP, LABEL) \
    __asm__("movl  $" STR(CONST) ",%edi");	/* edi = args_size */\
    bytepcoff = BYTEPCOFF;\
    \
    TGTOBJ_##VOP(LABEL, 0);			/* edx = obj */\
	/* TGTOBJ_* performs null check, too */\
    \
    METHODBLOCK_##VOP(LABEL);			/* eax = method */\
    \
    JITDO_INVOKE_##VOP(LABEL);\
    \
    VARSPACE_##VOP(LABEL);			/* ecx = local var. space */\
    \
    __asm__("subl  %ecx,%esp\n\t"\
	    "pushl %ecx\n\t"\
	    "leal  4(%esp,%ecx),%ecx");		/* ecx = original esp */\
    \
    __asm__("pushl %esi\n\tpushl %edi");	/* save */\
    \
    /*\
    __asm__("leal  -4(%ecx,%edi,4),%ecx");\
		/* ecx = original esp + (args_size * 4 - 1) */\
    \
    __asm__("pushl %ecx");			/* origianl esp */\
    __asm__("pushl %0" : : "m" (bytepcoff));	/* bytepcoff */\
    __asm__("pushl %eax\n\t"			/* method */\
	    "pushl %edx");			/* obj */\
    CALL_INVOKEMETHOD(LABEL)

#define CODE_INVOKE(vop, VOP, THROW_EXC) \
  CODE(opc_invoke##vop, invoke##vop, ST0, ST0, THROW_EXC) {\
    INVOKE_ST0(VOP, #vop "_st0");\
  }\
  CODE(opc_invoke##vop, invoke##vop, ST1, ST0, THROW_EXC) {\
    __asm__("pushl %edx");	/* now state 0 */\
    INVOKE_ST0(VOP, #vop "_st1");\
  }\
  CODE(opc_invoke##vop, invoke##vop, ST2, ST0, THROW_EXC) {\
    __asm__("pushl %edx\n\t"\
	    "pushl %ecx");	/* now state 0 */\
    INVOKE_ST0(VOP, #vop "_st2");\
  }\
  CODE(opc_invoke##vop, invoke##vop, ST3, ST0, THROW_EXC) {\
    __asm__("pushl %ecx");	/* now state 0 */\
    INVOKE_ST0(VOP, #vop "_st3");\
  }\
  CODE(opc_invoke##vop, invoke##vop, ST4, ST0, THROW_EXC) {\
    __asm__("pushl %ecx\n\t"\
	    "pushl %edx");	/* now state 0 */\
    INVOKE_ST0(VOP, #vop "_st4");\
  }


#define TGTOBJ_VIRTUAL(LABEL, STATE) \
    __asm__("movl  -4(%esp,%edi,4),%edx");\
    NULL_TEST(%edx, LABEL, STATE)

#ifdef JITDO
#  define METHODBLOCK_VIRTUAL(LABEL) \
    __asm__("pushl %edi");	/* save */\
    OBJ_ARRAY_METHODTABLE(%edx, %ecx, LABEL);	/* ecx = obj_array_..(obj) */\
    /* debug code */\
    \
    JUMP_IF_NOT_REMOTE(LABEL "_not_proxy");\
    \
    METHODTABLE_OF_PROXY(%edi);\
    __asm__("cmpl  %ecx,%edi\n\t"\
	    "jnz   " LABEL "_not_proxy");\
DEBUG_IN;\
PUSH_CONSTSTR("METDODBLOCK_VIRTUAL: Proxy is found.\n");\
__asm__("call printf@PLT\n\t"\
	"addl $4,%esp");\
FFLUSH;\
DEBUG_OUT;\
    PROXY_CLAZZ(%edx, %ecx);		/* ecx = Proxy.clazz */\
DEBUG_IN;\
__asm__("pushl %ecx");\
CB_NAME(%ecx, %ecx);\
__asm__("pushl %ecx");\
PUSH_CONSTSTR("  Proxy.clazz: %s 0x%08x\n");\
__asm__("call printf@PLT\n\t"\
	"addl $12,%esp");\
FFLUSH;\
DEBUG_OUT;\
    CB_METHODTABLE(%ecx, %ecx);		/* ecx = methodtable of Proxy.clazz*/\
DEBUG_IN;\
__asm__("pushl %ecx");\
PUSH_CONSTSTR("  methodtable: 0x%08x\n");\
__asm__("call printf@PLT\n\t"\
	"addl $8,%esp");\
FFLUSH;\
DEBUG_OUT;\
    \
    __asm__(LABEL "_not_proxy:");\
    __asm__("popl %edi");	/* restore */\
    __asm__("movl  $" STR(CONST) ",%eax");	/* eax = methodIndex */\
    MT_SLOT(%ecx, %eax, %eax)			/* eax = method */
#else
#define METHODBLOCK_VIRTUAL(LABEL) \
    /* get methodblock */\
    OBJ_ARRAY_METHODTABLE(%edx, %ecx, LABEL);	/* ecx = obj_array_..(obj) */\
    __asm__("movl  $" STR(CONST) ",%eax");	/* eax = methodIndex */\
    MT_SLOT(%ecx, %eax, %eax)			/* eax = method */
#endif	/* JITDO */

#define VARSPACE_VIRTUAL(LABEL) \
    /* allocate local var space */\
    METHOD_NLOCALS(%eax, %ecx);		/* ecx = method->nlocals */\
    __asm__("subl  %edi,%ecx\n\t"	/* ecx -= args_size */\
	    "jle   " LABEL "_nlocal_le_0\n\t"\
	    "shll  $2,%ecx\n\t"		/* ecx *= 4 */\
	    "jmp   " LABEL "_nlocal_done\n\t"\
	LABEL "_nlocal_le_0:\n\t"\
	    "xorl  %ecx,%ecx\n\t"	/* ecx = 0 */\
	LABEL "_nlocal_done:")

#define JITDO_INVOKE_VIRTUAL(LABEL)	JITDO_INVOKE(LABEL)

#define TGTOBJ_SPECIAL(LABEL, STATE)	TGTOBJ_VIRTUAL(LABEL, STATE)
#define METHODBLOCK_SPECIAL(LABEL)	__asm__("movl  $" STR(CONST) ",%eax")
#define VARSPACE_SPECIAL(LABEL)		__asm__("movl  $" STR(CONST) ",%ecx")
#define JITDO_INVOKE_SPECIAL(LABEL)	JITDO_INVOKE(LABEL)

#define TGTOBJ_STATIC(LABEL, STATE)
	/* null instead of 0 clear (__asm__("xorl  %edx,%edx")) */
#define METHODBLOCK_STATIC(LABEL) \
    __asm__("movl  $" STR(CONST) ",%eax");\
    METHOD_CLAZZ(%eax,%edx)	/* edx = cbHandle(method->fb.clazz) */
#define VARSPACE_STATIC(LABEL)		__asm__("movl  $" STR(CONST) ",%ecx")
#define JITDO_INVOKE_STATIC(LABEL)	/* nothing */

#define TGTOBJ_INTERFACE(LABEL, STATE)	TGTOBJ_VIRTUAL(LABEL, STATE)

#define METHODBLOCK_INTERFACE(LABEL) \
    /* get methodblock: call getInterfaceMethod() */\
    __asm__("pushl %edi\n\tpushl  %edx");\
    FUNCCALL_IN(0);\
    __asm__("pushl $" STR(CONST) "\n\t"	/* guessptr */\
	    "pushl $" STR(CONST) "\n\t"	/* imethod */\
	    "pushl %edx");		/* obj */\
    __asm__("pushl %0" : : "m" (ee));	/* ee */\
    CALL_GETINTFMETHOD;\
    FUNCCALL_OUT(0);\
    __asm__("popl  %edx\n\tpopl  %edi")

#define VARSPACE_INTERFACE(LABEL)	VARSPACE_VIRTUAL(LABEL)
#define JITDO_INVOKE_INTERFACE(LABEL)	JITDO_INVOKE(LABEL)

#ifdef RUNTIME_DEBUG
#  define CALL_GETINTFMETHOD \
    __asm__("pushl %0" : : "m" (runtime_debug));\
    __asm__("call  " SYMBOL(getInterfaceMethod) "@PLT\n\t"\
	    "addl  $20,%esp")
#else
#  define CALL_GETINTFMETHOD \
    __asm__("call  " SYMBOL(getInterfaceMethod) "@PLT\n\t"\
	    "addl  $16,%esp")
#endif


  CODE_INVOKE(virtual, VIRTUAL, 1);
  CODE_INVOKE(special, SPECIAL, 1);
  CODE_INVOKE(static, STATIC, 0);
  CODE_INVOKE(interface, INTERFACE, 1);

#if 0	/* how to get methodblock */
#  if 0
    if (opcode == opc_invokevirtual_quick)	/* obj isn't an array */
      method = mt_slot(obj_methodtable(obj), methodIndex);
    else
#  endif
      method = mt_slot(obj_array_methodtable(obj), methodIndex);
#endif
#if 0
__asm__("pushl %eax\n\t"\
	"movl  %esp,%eax\n\t"\
	"pusha\n\t"\
	"pushl %esp");\
PUSH_CONSTSTR("  ESP: 0x%08x\n");\
__asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	"addl  $8,%esp");\
FFLUSH;\
__asm__("popa\n\t"\
	"popl  %eax");
#endif

  /* invokevirtual_quick */
	/* const: args_size, methodIndex, retsize */
  CODE_INVOKE(virtual_quick, VIRTUAL, 1);

  /* invokevirtualobject_quick */
	/* const: args_size, methodIndex, retsize */
  CODE_INVOKE(virtualobject_quick, VIRTUAL, 1);

  /* xxxunusedxxx */
  CODE(opc_xxxunusedxxx, xxxunusesxxx, STANY, STSTA, 0) {}

  /* new */
	/* cosnt: cb */
#ifndef NO_CHECK
#  define NEW_TEST(CB_REG, LABEL, STATE) \
    /* call VerifyClassAccess() */\
    \
    __asm__("pushl " #CB_REG);\
    FUNCCALL_IN(STATE);\
    \
    __asm__("pushl $1\n\t"	/* classloader_only */\
	    "pushl " #CB_REG);	/* new_class */\
    /* %eax = ee->current_frame->current_method */\
    {\
      register ExecEnv *cur_ee asm("eax");\
      cur_ee = ee;\
    }\
    __asm__("movl  " EE_CURRENTFRAME(%eax) ",%edi\n\t"\
	    "movl  " FRAME_CURRENTMETHOD(%edi) ",%eax");\
    {\
      register struct methodblock *method asm("eax");\
      if (method)\
	__asm__("pushl %0" : : "m" (method->fb.clazz));\
      else\
	__asm__("pushl $0");\
    }\
    __asm__("call  " SYMBOL(VerifyClassAccess) "@PLT\n\t"\
	    "addl  $12,%esp");\
    FUNCCALL_OUT(STATE);\
    __asm__("popl  " #CB_REG);\
    \
    __asm__("testl %eax,%eax\n\t"\
	    "jnz   " LABEL "_testok");\
    SIGNAL_ERROR0(IllegalAccessError, STATE);\
    __asm__(LABEL "_testok:")
#else
#  define NEW_TEST(CB_REG, LABEL, STATE)
#endif

#ifdef RUNTIME_DEBUG
#  define NEW_DEBUG1(CB_REG) \
  if (runtime_debug) {\
    DEBUG_IN;\
    CB_NAME(CB_REG, %eax);  __asm__("pushl %eax");\
    PUSH_CONSTSTR("  clazz: %s\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $8,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#  define NEW_DEBUG2(OBJ_REG) \
  if (runtime_debug) {\
    DEBUG_IN;\
    __asm__("pushl " #OBJ_REG);\
    PUSH_CONSTSTR("  obj: 0x%08x\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	  "addl  $8,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#else
#  define NEW_DEBUG1(CB_REG)
#  define NEW_DEBUG2(OBJ_REG)
#endif

#ifdef JITDO
#  define JITDO_NEW(CB_REG, DST_REG, LABEL, STATE) \
    JUMP_IF_NOT_REMOTE(LABEL "_local");\
    \
    {	/* %edi = ee */\
      register ExecEnv *cur_ee asm("edi");\
      cur_ee = ee;\
    }\
    \
    /* local operation if remote VM addr is null */\
    __asm__("movl  " EE_REMOTE_ADDR(%edi) ",%eax\n\t"\
	    "testl %eax,%eax\n\t"\
	    "jz    " LABEL "_local");\
    \
    /* call isCopyable() */\
    __asm__("pushl %edi");\
    FUNCCALL_IN(STATE);\
    __asm__("pushl " #CB_REG "\n\t"\
	    "pushl %edi\n\t"\
	    "call  " SYMBOL(isCopyable) "@PLT");\
    __asm__("addl  $4,%esp\n\t"\
	    "popl  " #CB_REG);\
    FUNCCALL_OUT(STATE);\
    __asm__("popl  %edi");\
    __asm__("testl %eax,%eax\n\t"\
	    "jnz   " LABEL "_local");\
    \
    /* call proxy_new() */\
    FUNCCALL_IN(STATE);\
    __asm__("pushl " #CB_REG "\n\t"\
	    "pushl " EE_REMOTE_ADDR(%edi) "\n\t"\
	    "pushl %edi\n\t"	/* ee */\
	    "call  " SYMBOL(proxy_new) "@PLT\n\t"\
	    "addl  $12,%esp");\
    FUNCCALL_OUT(STATE);\
    \
    JUMP_IF_EXC_HASNT_OCCURRED(%edi /* is ee */, LABEL "_done");\
    SIGNAL_ERROR_JUMP(STATE);\
    __asm__("jmp   " LABEL "_done\n\t"\
	    LABEL "_local:")
#else
#  define JITDO_NEW(CB_REG, DST_REG, LABEL, STATE)
#endif	/* JITDO */

/* stuff to rewrite code */
#if defined(linux)
#  ifdef RUNTIME_DEBUG
#    define NEW_REWRITE_OFFSET_0	0x32
#  else
#    define NEW_REWRITE_OFFSET_0	5
#  endif
#elif defined(__FreeBSD__)
#  ifdef RUNTIME_DEBUG
#    define NEW_REWRITE_OFFSET_0	0x37
#  else
#    define NEW_REWRITE_OFFSET_0	5
#  endif
#endif

#  define NEW_REWRITE_OFFSET_1	NEW_REWRITE_OFFSET_0
#  define NEW_REWRITE_OFFSET_2	(NEW_REWRITE_OFFSET_0 + 1)
#  define NEW_REWRITE_OFFSET_3	NEW_REWRITE_OFFSET_0
#  define NEW_REWRITE_OFFSET_4	(NEW_REWRITE_OFFSET_0 + 1)

#ifndef NO_REWRITE_NEW
#  define QUICK_NEW(CB_REG, LABEL) \
    __asm__("jmp   " LABEL "_test\n\t"\
	    "jmp   " LABEL "_testdone\n\t"\
	LABEL "_test:")
#  define REWRITE_NEW(REWRITE_OFFSET) \
    /* rewrite self */\
    COMPILEDCODE(%eax);\
    __asm__("addl  $" STR(CONST) ",%eax");\
    __asm__("movb  $0x90," STR(REWRITE_OFFSET) "(%eax)\n\t"\
	    "movb  $0x90," STR(REWRITE_OFFSET + 1) "(%eax)")
#else
#  define QUICK_NEW(CB_REG, LABEL)
#  define REWRITE_NEW(REWRITE_OFFSET)
#endif

#define NEW(CB_REG, DST_REG, REWRITE_OFFSET, LABEL, STATE) \
    __asm__("movl  $" STR(CONST) "," #CB_REG);	/* cb */\
    \
    QUICK_NEW(CB_REG, LABEL);\
    \
    CB_ACCESS(CB_REG, %eax);\
    __asm__("testb $6,%al\n\t"\
	    "jz    " LABEL);\
    SIGNAL_ERROR0(InstantiationError, STATE);\
    __asm__(LABEL ":");\
    \
    NEW_TEST(CB_REG, LABEL, STATE);\
    \
    REWRITE_NEW(REWRITE_OFFSET);\
    \
    __asm__(LABEL "_testdone:");\
    \
    NEW_DEBUG1(CB_REG);\
    \
    /* instantiate */\
    JITDO_NEW(CB_REG, DST_REG, LABEL, STATE);\
    /* call newobject() */\
    FUNCCALL_IN(STATE);\
    __asm__("pushl %0" : : "m" (ee));	/* ee */\
    __asm__("pushl $0\n\t"		/* pc */\
	    "pushl " #CB_REG "\n\t"\
	    "call  " SYMBOL(newobject) "@PLT\n\t"\
	    "addl  $12,%esp");\
    FUNCCALL_OUT(STATE);\
    NEW_DEBUG2(%eax);\
    __asm__(LABEL "_done:");\
    __asm__("movl  %eax," #DST_REG)


  CODE(opc_new, new, ST0, ST1, 0) {
    NEW(%edx, %edx, NEW_REWRITE_OFFSET_0, "new_st0", 0);
  }
  CODE(opc_new, new, ST1, ST2, 0) {
    NEW(%ecx, %ecx, NEW_REWRITE_OFFSET_1, "new_st1", 1);
  }
  CODE(opc_new, new, ST2, ST4, 0) {
    __asm__("pushl %edx");	/* now state 3 */
    NEW(%edx, %edx, NEW_REWRITE_OFFSET_2, "new_st2", 3);
  }
  CODE(opc_new, new, ST3, ST4, 0) {
    NEW(%edx, %edx, NEW_REWRITE_OFFSET_3, "new_st3", 3);
  }
  CODE(opc_new, new, ST4, ST2, 0) {
    __asm__("pushl %ecx");	/* now state 1 */
    NEW(%ecx, %ecx, NEW_REWRITE_OFFSET_4, "new_st4", 1);
  }

  /* newarray */
	/* const: type */
#ifndef NO_CHECK
#  define NEWARRAY_TEST(OPTOP1_REG, LABEL, STATE) \
    __asm__("testl " #OPTOP1_REG "," #OPTOP1_REG "\n\t"\
	    "jge   " LABEL "_1");\
    SIGNAL_ERROR0(NegativeArraySizeException, STATE);\
    __asm__(LABEL "_1:")
#else
#  define NEWARRAY_TEST(OPTOP1_REG, LABEL, STATE)
#endif

#ifdef RUNTIME_DEBUG
#  define NEWARRAY_DEBUG1(TYPE, COUNT) \
  if (runtime_debug) {\
    DEBUG_IN;\
    __asm__("pushl " #COUNT "\n\tpushl " #TYPE);\
    PUSH_CONSTSTR("  type: 0x%x, count: %d\n");\
    __asm__("call  " SYMBOL(printf) "@PLT\n\t"\
	    "addl  $12,%esp");\
    FFLUSH;\
    DEBUG_OUT;\
  }
#else
#  define NEWARRAY_DEBUG1(TYPE, COUNT)
#endif

#define NEWARRAY(OPTOP1_REG, LABEL, STATE) \
    __asm__("movl  $" STR(CONST) ",%eax");	/* eax = type */\
    NEWARRAY_DEBUG1(%eax, OPTOP1_REG);\
    \
    NEWARRAY_TEST(OPTOP1_REG, LABEL, STATE);\
    \
    /* call ArrayAlloc() */\
    FUNCCALL_IN(STATE);\
    __asm__("pushl " #OPTOP1_REG "\n\t"	/* count */\
	    "pushl %eax\n\t"		/* type */\
	    "call  " SYMBOL(ArrayAlloc) "@PLT\n\t"\
	    "addl  $8,%esp");\
    FUNCCALL_OUT(STATE);\
    __asm__("testl %eax,%eax\n\t"\
	    "jnz   " LABEL "_done");\
    SIGNAL_ERROR0(OutOfMemoryError, STATE);\
    __asm__(LABEL "_done:\n\t"\
	    "movl  %eax," #OPTOP1_REG)	/* store to dst. */

  CODE(opc_newarray, newarray, ST0, ST1, 0) {
    __asm__("popl  %edx");	/* now state 1 */
    NEWARRAY(%edx, "newarray_st0", 1);
  }
  CODE(opc_newarray, newarray, ST1, ST1, 0) {
    NEWARRAY(%edx, "newarray_st1", 1);
  }
  CODE(opc_newarray, newarray, ST2, ST2, 0) {
    NEWARRAY(%ecx, "newarray_st2", 2);
  }
  CODE(opc_newarray, newarray, ST3, ST3, 0) {
    NEWARRAY(%ecx, "newarray_st3", 3);
  }
  CODE(opc_newarray, newarray, ST4, ST4, 0) {
    NEWARRAY(%edx, "newarray_st4", 4);
  }

  /* anewarray */
	/* const: elem_clazz */
#define ANEWARRAY(OPTOP1_REG, LABEL, STATE) \
    NEWARRAY_TEST(OPTOP1_REG, LABEL, STATE);\
    \
    /* call ArrayAlloc() */\
    FUNCCALL_IN(STATE);\
    __asm__("pushl " #OPTOP1_REG "\n\t"			/* count */\
	    "pushl $" STR(T_CLASS) "\n\t"	/* type (T_CLASS) */\
	    "call  " SYMBOL(ArrayAlloc) "@PLT\n\t"\
	    "addl  $8,%esp");\
    FUNCCALL_OUT(STATE);\
    __asm__("testl %eax,%eax\n\t"\
	    "jnz   " LABEL "_2");\
    SIGNAL_ERROR0(OutOfMemoryError, STATE);\
    __asm__(LABEL "_2:");\
    \
    UNHAND(%eax, %edi);\
    __asm__("movl  $" STR(CONST) ",(%edi," #OPTOP1_REG ",4)");\
	/* unhand(array)->body[count] = clazz of elements */\
    __asm__("movl  %eax," #OPTOP1_REG)		/* store to dst. */

  CODE(opc_anewarray, anewarray, ST0, ST1, 0) {
    __asm__("popl  %edx");	/* now state 1 */
    ANEWARRAY(%edx, "anewarray_st0", 1);
  }
  CODE(opc_anewarray, anewarray, ST1, ST1, 0) {
    ANEWARRAY(%edx, "anewarray_st1", 1);
  }
  CODE(opc_anewarray, anewarray, ST2, ST2, 0) {
    ANEWARRAY(%ecx, "anewarray_st2", 2);
  }
  CODE(opc_anewarray, anewarray, ST3, ST3, 0) {
    ANEWARRAY(%ecx, "anewarray_st3", 3);
  }
  CODE(opc_anewarray, anewarray, ST4, ST4, 0) {
    ANEWARRAY(%edx, "anewarray_st4", 4);
  }

  /* arraylength */
  CODE(opc_arraylength, arraylength, ST0, ST3, 1) {
    __asm__("popl  %edx");	/* now state 1 */
    NULL_TEST(%edx, "arylen_null_st0", 0);
    OBJ_LENGTH(%edx, %ecx);
  }
  CODE(opc_arraylength, arraylength, ST1, ST3, 1) {
    NULL_TEST(%edx, "arylen_null_st1", 1);
    OBJ_LENGTH(%edx, %ecx);
  }
  CODE(opc_arraylength, arraylength, ST2, ST2, 1) {
    NULL_TEST(%ecx, "arylen_null_st2", 2);
    OBJ_LENGTH(%ecx, %ecx);
  }
  CODE(opc_arraylength, arraylength, ST3, ST1, 1) {
    NULL_TEST(%ecx, "arylen_null_st3", 3);
    OBJ_LENGTH(%ecx, %edx);
  }
  CODE(opc_arraylength, arraylength, ST4, ST4, 1) {
    NULL_TEST(%edx, "arylen_null_st4", 4);
    OBJ_LENGTH(%edx, %edx);
  }

  /* athrow */
#define ATHROW(OPTOP1_REG, STATE) \
    NULL_TEST_CANT_BE_ELIMINATED(OPTOP1_REG, "athrow_st" #STATE "_1", STATE);\
    \
    /* macro exceptionThrow(ee, obj) */\
    {\
      register ExecEnv *cur_ee asm("eax");\
      register JHandle *obj asm(#OPTOP1_REG);\
      \
      cur_ee = ee;\
      cur_ee->exceptionKind = EXCKIND_THROW;\
      cur_ee->exception.exc = obj;\
    }\
    \
    bytepcoff = BYTEPCOFF;\
    __asm__("jmp  " STR(ADDR_EXC))

  CODE(opc_athrow, athrow, ST0, ST1, 1) {
    __asm__("popl  %edx");	/* now state 1 */
    ATHROW(%edx, 0);
  }
  CODE(opc_athrow, athrow, ST1, ST1, 1) {
    ATHROW(%edx, 1);
  }
  CODE(opc_athrow, athrow, ST2, ST2, 1) {
    ATHROW(%ecx, 2);
  }
  CODE(opc_athrow, athrow, ST3, ST3, 1) {
    ATHROW(%ecx, 3);
  }
  CODE(opc_athrow, athrow, ST4, ST4, 1) {
    ATHROW(%edx, 4);
  }

  /* checkcast */
	/* const: cb */
#ifdef JITDO
#  define JITDO_CHECKCAST(OPTOP1_REG, LABEL, STATE) \
    /* compare handle->methods with cb_of_Proxy->methodtable */\
    JUMP_IF_NOT_REMOTE(LABEL "_local");\
    JUMP_IF_NOT_PROXY(%##OPTOP1_REG, LABEL "_local");\
    {\
      register HNET_shudo_jitdo_Proxy *h asm(#OPTOP1_REG);\
      register ClassNET_shudo_jitdo_Proxy *edi asm("edi");\
      register Hjava_lang_Class *clz asm("eax");\
      \
      edi = unhand(h);\
      clz = edi->clazz;\
    }	/* eax = Proxy.clazz */\
    __asm__("movl $" STR(CONST) ",%edi\n\t" /* cb */\
	    "cmpl %eax,%edi\n\t"\
	    "jnz  " LABEL "_fail\n\t"\
	    "jmp  " LABEL "_done\n\t"\
	    LABEL "_local:");
#else
#  define JITDO_CHECKCAST(OPTOP1_REG, LABEL, STATE)
#endif	/* JITDO */

#define CHECKCAST(OPTOP1_REG, LABEL, STATE) \
    __asm__("testl %" #OPTOP1_REG ",%" #OPTOP1_REG "\n\t"\
	    "jz    " LABEL "_done");\
    JITDO_CHECKCAST(OPTOP1_REG, LABEL, STATE);\
    FUNCCALL_IN(STATE);\
    __asm__("pushl %0" : : "m" (ee));	/* ee */\
    __asm__("pushl $" STR(CONST) "\n\t"	/* cb */\
	    "pushl %" #OPTOP1_REG "\n\t"\
	    "call  " SYMBOL(is_instance_of) "@PLT\n\t"\
	    "addl  $12,%esp");\
    FUNCCALL_OUT(STATE);\
    __asm__("testl %eax,%eax\n\t"\
	    "jnz   " LABEL "_done");\
    __asm__(LABEL "_fail:");\
    SIGNAL_ERROR0(ClassCastException, STATE);\
    __asm__(LABEL "_done:")

  CODE(opc_checkcast, checkcast, ST0, ST1, 0) {
    __asm__("popl  %edx");	/* now state 1 */
    CHECKCAST(edx, "checkcast_st0", 1);
  }
  CODE(opc_checkcast, checkcast, ST1, ST1, 0) {
    CHECKCAST(edx, "checkcast_st1", 1);
  }
  CODE(opc_checkcast, checkcast, ST2, ST2, 0) {
    CHECKCAST(ecx, "checkcast_st2", 2);
  }
  CODE(opc_checkcast, checkcast, ST3, ST3, 0) {
    CHECKCAST(ecx, "checkcast_st3", 3);
  }
  CODE(opc_checkcast, checkcast, ST4, ST4, 0) {
    CHECKCAST(edx, "checkcast_st4", 4);
  }

  /* instanceof */
	/* const: cb */
#ifdef JITDO
#  define JITDO_INSTANCEOF(OPTOP1_REG, LABEL, STATE) \
    /* compare handle->methods with cb_of_Proxy->methodtable */\
    JUMP_IF_NOT_PROXY(%##OPTOP1_REG, LABEL "_local");\
    {\
      register HNET_shudo_jitdo_Proxy *h asm(#OPTOP1_REG);\
      register ClassNET_shudo_jitdo_Proxy *edi asm("edi");\
      register Hjava_lang_Class *clz asm("eax");\
      \
      edi = unhand(h);\
      clz = edi->clazz;\
    }	/* eax = Proxy.clazz */\
    __asm__("movl $" STR(CONST) ",%edi\n\t" /* cb */\
	    "cmpl %eax,%edi\n\t"\
	    "jz   " LABEL "_false\n\t"\
	    "movl $1,%" #OPTOP1_REG "\n\t"\
	    "jmp  " LABEL "_done\n\t"\
	    LABEL "_false:"\
	    "movl $0,%" #OPTOP1_REG "\n\t"\
	    "jmp  " LABEL "_done\n\t"\
	    LABEL "_local:");
#else
#  define JITDO_INSTANCEOF(OPTOP1_REG, LABEL, STATE)
#endif	/* JITDO */

#define INSTANCEOF(OPTOP1_REG, LABEL, STATE) \
    __asm__("testl %" #OPTOP1_REG ",%" #OPTOP1_REG "\n\t"\
	    "jz    " LABEL "_done");\
    JITDO_INSTANCEOF(OPTOP1_REG, LABEL, STATE);\
    FUNCCALL_IN(STATE);\
    __asm__("pushl %0" : : "m" (ee));	/* ee */\
    __asm__("pushl $" STR(CONST) "\n\t"\
	    "pushl %" #OPTOP1_REG "\n\t"\
	    "call  " SYMBOL(is_instance_of) "@PLT\n\t"\
	    "addl  $12,%esp");\
    FUNCCALL_OUT(STATE);\
    __asm__("movl  %eax,%" #OPTOP1_REG "\n\t"\
	LABEL "_done:")

  CODE(opc_instanceof, instanceof, ST0, ST1, 0) {
    __asm__("popl  %edx");	/* now state 1 */
    INSTANCEOF(edx, "instanceof_st0", 1);
  }
  CODE(opc_instanceof, instanceof, ST1, ST1, 0) {
    INSTANCEOF(edx, "instanceof_st1", 1);
  }
  CODE(opc_instanceof, instanceof, ST2, ST2, 0) {
    INSTANCEOF(ecx, "instanceof_st2", 2);
  }
  CODE(opc_instanceof, instanceof, ST3, ST3, 0) {
    INSTANCEOF(ecx, "instanceof_st3", 3);
  }
  CODE(opc_instanceof, instanceof, ST4, ST4, 0) {
    INSTANCEOF(edx, "instanceof_st4", 4);
  }

  /* monitorenter, monitorexit */
#define MONITOR(OPTOP1_REG, FUNCNAME, LABEL, STATE) \
    NULL_TEST_CANT_BE_ELIMINATED(OPTOP1_REG, LABEL "_1", STATE);\
    OBJ_MONITOR(OPTOP1_REG, OPTOP1_REG);\
    FUNCCALL_IN(STATE);\
    __asm__("pushl " #OPTOP1_REG "\n\t"\
	    "call  " #FUNCNAME "@PLT\n\t"\
	    "addl  $4,%esp");\
    FUNCCALL_OUT(STATE)

#define CODE_MONITOR(vop, FUNCNAME) \
  CODE(opc_##vop, ##vop, ST0, ST0, 1) {\
    __asm__("popl  %edx");	/* now state 1 */\
    MONITOR(%edx, FUNCNAME, #vop "_st0", 0);\
  }\
  CODE(opc_##vop, ##vop, ST1, ST0, 1) {\
    MONITOR(%edx, FUNCNAME, #vop "_st1", 0);\
  }\
  CODE(opc_##vop, ##vop, ST2, ST1, 1) {\
    MONITOR(%ecx, FUNCNAME, #vop "_st2", 1);\
  }\
  CODE(opc_##vop, ##vop, ST3, ST0, 1) {\
    MONITOR(%ecx, FUNCNAME, #vop "_st3", 0);\
  }\
  CODE(opc_##vop, ##vop, ST4, ST3, 1) {\
    MONITOR(%edx, FUNCNAME, #vop "_st4", 3);\
  }

  CODE_MONITOR(monitorenter, monitorEnter);
  CODE_MONITOR(monitorexit, monitorExit);

  /* multianewarray */
	/* const: dimensions, arrayclazz */
#define MULTIANEWARRAY_TEST(STATE) \
    __asm__("cmpl  $-1,%eax\n\t"  /* eax is returned by multianewarray() */\
	    "jne   mulary_st" #STATE "_1");\
    SIGNAL_ERROR0(NegativeArraySizeException, STATE);\
    __asm__("mulary_st" #STATE "_1:");\
    __asm__("testl %eax,%eax\n\t"\
	    "jnz   mulary_st" #STATE "_2");\
    SIGNAL_ERROR0(OutOfMemoryError, STATE);\
    __asm__("mulary_st" #STATE "_2:")

#ifdef RUNTIME_DEBUG
#  define MULTIANEWARRAY_FUNC \
  __asm__("pushl %0" : : "m" (runtime_debug));\
  __asm__("call  " SYMBOL(multianewarray) "@PLT\n\t"\
	  "addl  $20,%esp")
#else
#  define MULTIANEWARRAY_FUNC \
  __asm__("call  " SYMBOL(multianewarray) "@PLT\n\t"\
	  "addl  $16,%esp")
#endif

#define MULTIANEWARRAY_ST0(DST_REG, STATE) \
    __asm__("movl  $" STR(CONST) ",%edi");	/* dimensions */\
    \
    __asm__("movl  %esp,%eax");\
    \
    FUNCCALL_IN(0);\
    __asm__("pushl %edi");\
    __asm__("pushl %eax\n\t"\
	    "pushl $" STR(CONST) "\n\t"	/* arrayclazz */\
	    "pushl %edi");		/* dimensions */\
    __asm__("pushl %0" : : "m" (ee));	/* ee */\
    MULTIANEWARRAY_FUNC;\
    __asm__("popl  %edi");\
    FUNCCALL_OUT(0);\
    \
    __asm__("leal  (%esp,%edi,4),%esp");/* pop dimensions */\
    \
    MULTIANEWARRAY_TEST(STATE);\
    __asm__("movl  %eax," #DST_REG)

  CODE(opc_multianewarray, multianewarray, ST0, ST1, 0) {
    MULTIANEWARRAY_ST0(%edx, 0);
  }
  CODE(opc_multianewarray, multianewarray, ST1, ST1, 0) {
    __asm__("pushl %edx");	/* now state 0 */
    MULTIANEWARRAY_ST0(%edx, 1);
  }
  CODE(opc_multianewarray, multianewarray, ST2, ST1, 0) {
    __asm__("pushl %edx\n\t"
	    "pushl %ecx");	/* now state 0 */
    MULTIANEWARRAY_ST0(%edx, 2);
  }
  CODE(opc_multianewarray, multianewarray, ST3, ST1, 0) {
    __asm__("pushl %ecx");	/* now state 0 */
    MULTIANEWARRAY_ST0(%edx, 3);
  }
  CODE(opc_multianewarray, multianewarray, ST4, ST1, 0) {
    __asm__("pushl %ecx\n\t"
	    "pushl %edx");	/* now state 0 */
    MULTIANEWARRAY_ST0(%edx, 4);
  }

  /* invokeignored_quick */
	/* const: args_size */
#define INVOKEIGNORED_QUICK_ST0(STATE) \
    __asm__("movl  $" STR(CONST) ",%edi\n\t"\
	    "movl  -4(%esp,%edi,4),%eax\n\t"\
	    "leal  (%esp,%edi,4),%esp\n\t"\
	    "testl %eax,%eax\n\t"\
	    "jnz   invign_st" #STATE "_done");\
    SIGNAL_ERROR0(NullPointerException, STATE);\
    __asm__("invign_st" #STATE "_done:")

#define INVOKEIGNORED_NOCHECK_ST0(STATE) \
    __asm__("movl  $" STR(CONST) ",%edi\n\t"\
	    "leal  (%esp,%edi,4),%esp")

#define CODE_INVOKEIGNORED(suffix, SUFFIX) \
  CODE(opc_invokeignored_##suffix, invokeignored_##suffix, ST0, ST0, 0) {\
    INVOKEIGNORED_##SUFFIX##_ST0(0);\
  }\
  CODE(opc_invokeignored_##suffix, invokeignored_##suffix, ST1, ST0, 0) {\
    __asm__("pushl %edx");	/* now state 0 */\
    INVOKEIGNORED_##SUFFIX##_ST0(1);\
  }\
  CODE(opc_invokeignored_##suffix, invokeignored_##suffix, ST2, ST0, 0) {\
    __asm__("pushl %edx\n\t"\
	    "pushl %ecx");	/* now state 0 */\
    INVOKEIGNORED_##SUFFIX##_ST0(2);\
  }\
  CODE(opc_invokeignored_##suffix, invokeignored_##suffix, ST3, ST0, 0) {\
    __asm__("pushl %ecx");	/* now state 0 */\
    INVOKEIGNORED_##SUFFIX##_ST0(3);\
  }\
  CODE(opc_invokeignored_##suffix, invokeignored_##suffix, ST4, ST0, 0) {\
    __asm__("pushl %ecx\n\t"\
	    "pushl %edx");	/* now state 0 */\
    INVOKEIGNORED_##SUFFIX##_ST0(4);\
  }

  CODE_INVOKEIGNORED(quick, QUICK);
  CODE_INVOKEIGNORED(nocheck, NOCHECK);

  /* new_quick */
	/* const: cb */
#define NEW_QUICK(DST_REG, STATE) \
    FUNCCALL_IN(STATE);\
    __asm__("pushl %0" : : "m" (ee));	/* ee */\
    __asm__("pushl $0\n\t"		/* pc */\
	    "pushl $" STR(CONST) "\n\t"\
	    "call  " SYMBOL(newobject) "@PLT\n\t"\
	    "addl  $12,%esp");\
    FUNCCALL_OUT(STATE);\
    __asm__("movl  %eax," #DST_REG)

  CODE(opc_new_quick, new_quick, ST0, ST1, 0) {
    NEW_QUICK(%edx, 0);
  }
  CODE(opc_new_quick, new_quick, ST1, ST2, 0) {
    NEW_QUICK(%ecx, 1);
  }
  CODE(opc_new_quick, new_quick, ST2, ST4, 0) {
    __asm__("pushl %edx");	/* now state 3 */
    NEW_QUICK(%edx, 3);
  }
  CODE(opc_new_quick, new_quick, ST3, ST4, 0) {
    NEW_QUICK(%edx, 3);
  }
  CODE(opc_new_quick, new_quick, ST4, ST2, 0) {
    __asm__("pushl %ecx");	/* now state 1 */
    NEW_QUICK(%ecx, 1);
  }

  /* nonnull_quick */
  CODE(opc_nonnull_quick, nonnull_quick, ST0, ST0, 0) {
    __asm__("popl  %edx");	/* now state 1 */
    NULL_TEST(%edx, "nonnull_quick_st0_1", 0);
  }
  CODE(opc_nonnull_quick, nonnull_quick, ST1, ST0, 0) {
    NULL_TEST(%edx, "nonnull_quick_st1_1", 1);
  }
  CODE(opc_nonnull_quick, nonnull_quick, ST2, ST1, 0) {
    NULL_TEST(%ecx, "nonnull_quick_st2_1", 2);
  }
  CODE(opc_nonnull_quick, nonnull_quick, ST3, ST0, 0) {
    NULL_TEST(%ecx, "nonnull_quick_st3_1", 3);
  }
  CODE(opc_nonnull_quick, nonnull_quick, ST4, ST3, 0) {
    NULL_TEST(%edx, "nonnull_quick_st4_1", 4);
  }

  CODEEND;
}
